from typing import Annotated, Any

import httpx
from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import X
from arcade_tdk.errors import RetryableToolError

from arcade_x.tools.constants import TWEETS_URL
from arcade_x.tools.utils import (
    expand_attached_media,
    expand_long_tweet,
    expand_urls_in_tweets,
    get_headers_with_token,
    get_tweet_url,
    parse_search_recent_tweets_response,
    remove_none_values,
)

# Manage Tweets Tools. See developer docs for additional available parameters:
# https://developer.x.com/en/docs/x-api/tweets/manage-tweets/api-reference


@tool(
    requires_auth=X(
        scopes=["tweet.read", "tweet.write", "users.read"],
    )
)
async def post_tweet(
    context: ToolContext,
    tweet_text: Annotated[str, "The text content of the tweet you want to post"],
) -> Annotated[str, "Success string and the URL of the tweet"]:
    """Post a tweet to X (Twitter)."""

    headers = get_headers_with_token(context)
    payload = {"text": tweet_text}

    async with httpx.AsyncClient() as client:
        response = await client.post(TWEETS_URL, headers=headers, json=payload, timeout=10)
        response.raise_for_status()

    tweet_id = response.json()["data"]["id"]
    return f"Tweet with id {tweet_id} posted successfully. URL: {get_tweet_url(tweet_id)}"


@tool(requires_auth=X(scopes=["tweet.read", "tweet.write", "users.read"]))
async def delete_tweet_by_id(
    context: ToolContext,
    tweet_id: Annotated[str, "The ID of the tweet you want to delete"],
) -> Annotated[str, "Success string confirming the tweet deletion"]:
    """Delete a tweet on X (Twitter)."""

    headers = get_headers_with_token(context)
    url = f"{TWEETS_URL}/{tweet_id}"

    async with httpx.AsyncClient() as client:
        response = await client.delete(url, headers=headers, timeout=10)
        response.raise_for_status()

    return f"Tweet with id {tweet_id} deleted successfully."


@tool(requires_auth=X(scopes=["tweet.read", "users.read"]))
async def search_recent_tweets_by_username(
    context: ToolContext,
    username: Annotated[str, "The username of the X (Twitter) user to look up"],
    max_results: Annotated[
        int, "The maximum number of results to return. Must be in range [1, 100] inclusive"
    ] = 10,
    next_token: Annotated[
        str | None, "The pagination token starting from which to return results"
    ] = None,
) -> Annotated[dict[str, Any], "Dictionary containing the search results"]:
    """Search for recent tweets (last 7 days) on X (Twitter) by username.
    Includes replies and reposts."""

    headers = get_headers_with_token(context)
    params: dict[str, Any] = {
        "query": f"from:{username}",
        "max_results": min(
            max(max_results, 10), 100
        ),  # X API does not allow 'max_results' less than 10 or greater than 100
        "next_token": next_token,
        "expansions": "author_id",
        "user.fields": "id,name,username,entities",
        "tweet.fields": "entities,note_tweet",
    }
    params = expand_attached_media(remove_none_values(params))

    url = f"{TWEETS_URL}/search/recent"

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, params=params, timeout=10)
        response.raise_for_status()

    response_data: dict[str, Any] = response.json()

    for tweet in response_data.get("data", []):
        expand_long_tweet(tweet)

    # Expand the URLs that are in the tweets
    response_data["data"] = expand_urls_in_tweets(
        response_data.get("data", []), delete_entities=True
    )

    # Parse the response data
    response_data = parse_search_recent_tweets_response(response_data)

    return response_data


@tool(requires_auth=X(scopes=["tweet.read", "users.read"]))
async def search_recent_tweets_by_keywords(
    context: ToolContext,
    keywords: Annotated[
        list[str] | None, "List of keywords that must be present in the tweet"
    ] = None,
    phrases: Annotated[
        list[str] | None, "List of phrases that must be present in the tweet"
    ] = None,
    max_results: Annotated[
        int, "The maximum number of results to return. Must be in range [1, 100] inclusive"
    ] = 10,
    next_token: Annotated[
        str | None, "The pagination token starting from which to return results"
    ] = None,
) -> Annotated[dict[str, Any], "Dictionary containing the search results"]:
    """
    Search for recent tweets (last 7 days) on X (Twitter) by required keywords and phrases.
    Includes replies and reposts.
    One of the following input parameters MUST be provided: keywords, phrases
    """

    if not any([keywords, phrases]):
        raise RetryableToolError(
            "No keywords or phrases provided",
            developer_message="Predicted inputs didn't contain any keywords or phrases",
            additional_prompt_content="Please provide at least one keyword or phrase for search",
            retry_after_ms=500,  # Play nice with X API rate limits
        )

    headers = get_headers_with_token(context)

    query = "".join([f'"{phrase}" ' for phrase in (phrases or [])])
    if keywords:
        query += " ".join(keywords or [])

    params: dict[str, Any] = {
        "query": query.strip(),
        "max_results": min(
            max(max_results, 10), 100
        ),  # X API does not allow 'max_results' less than 10 or greater than 100
        "next_token": next_token,
        "expansions": "author_id",
        "user.fields": "id,name,username,entities",
        "tweet.fields": "entities,note_tweet",
    }
    params = expand_attached_media(remove_none_values(params))

    url = f"{TWEETS_URL}/search/recent"

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, params=params, timeout=10)
        response.raise_for_status()

    response_data: dict[str, Any] = response.json()

    for tweet in response_data.get("data", []):
        expand_long_tweet(tweet)

    # Expand the URLs that are in the tweets
    response_data["data"] = expand_urls_in_tweets(
        response_data.get("data", []), delete_entities=True
    )

    # Parse the response data
    response_data = parse_search_recent_tweets_response(response_data)

    return response_data


@tool(requires_auth=X(scopes=["tweet.read", "users.read"]))
async def lookup_tweet_by_id(
    context: ToolContext,
    tweet_id: Annotated[str, "The ID of the tweet you want to look up"],
) -> Annotated[dict[str, Any], "Dictionary containing the tweet data"]:
    """Look up a tweet on X (Twitter) by tweet ID."""

    headers = get_headers_with_token(context)
    params = {
        "expansions": "author_id",
        "user.fields": "id,name,username,entities",
        "tweet.fields": "entities,note_tweet",
    }
    params = expand_attached_media(params)

    url = f"{TWEETS_URL}/{tweet_id}"

    async with httpx.AsyncClient() as client:
        response = await client.get(url, headers=headers, params=params, timeout=10)
        response.raise_for_status()

    response_data: dict[str, Any] = response.json()

    # Get the tweet data
    tweet_data = response_data.get("data")
    if tweet_data:
        expand_long_tweet(tweet_data)

        # Expand the URLs that are in the tweet
        expanded_tweet_list = expand_urls_in_tweets([tweet_data], delete_entities=True)
        response_data["data"] = expanded_tweet_list[0]
    else:
        response_data["data"] = {}

    return response_data



from arcade_tdk import ToolCatalog
from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_evals.critic import SimilarityCritic

import {{ package_name }}
from {{ package_name }}.tools.hello import say_hello

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module({{ package_name }})


@tool_eval()
def {{ toolkit_name }}_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="{{ toolkit_name }} Tools Evaluation",
        system_message=(
            "You are an AI assistant with access to {{ toolkit_name }} tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Saying hello",
        user_message="He's actually right here, say hi to him!",
        expected_tool_calls=[ExpectedToolCall(func=say_hello, args={"name": "John Doe"})],
        rubric=rubric,
        critics=[
            SimilarityCritic(critic_field="name", weight=0.5),
        ],
        additional_messages=[
            {"role": "user", "content": "My friend's name is John Doe."},
            {"role": "assistant", "content": "It is great that you have a friend named John Doe!"},
        ],
    )

    return suite




from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_salesforce
from arcade_salesforce.tools import get_account_data_by_id, get_account_data_by_keywords

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_salesforce)


@tool_eval()
def get_account_data_by_keywords_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="get_account_data_by_keywords",
        system_message=(
            "You are an AI assistant with access to Salesforce tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get account data by keywords",
        user_message="Get information about the Acme account.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get account data by keywords with limit",
        user_message="Get information of up to 3 accounts with the word 'Acme' in their name.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "limit": 3,
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=1 / 3),
            BinaryCritic(critic_field="limit", weight=1 / 3),
            BinaryCritic(critic_field="page", weight=1 / 3),
        ],
    )

    suite.add_case(
        name="Get summary of account activity",
        user_message="Get a summary of the latest activities in the Acme account.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    suite.add_case(
        name="Interactions with contacts of an account",
        user_message="Get me the interactions with the contacts of the Acme account.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    suite.add_case(
        name="Emails or calls with contacts of an account",
        user_message="Were there any emails or calls with contacts of the Acme account this week?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get account status",
        user_message="What's the status of the Acme account?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get overdue tasks",
        user_message="Are there any tasks overdue for the Acme account?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    suite.add_case(
        name="Get account closing likelihood",
        user_message="What's the likelihood of the Acme account closing a deal?",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_keywords,
                args={
                    "query": "Acme",
                    "page": 1,
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="query", weight=0.5),
            BinaryCritic(critic_field="page", weight=0.5),
        ],
    )

    return suite


@tool_eval()
def get_account_data_by_id_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="get_account_data_by_id",
        system_message=(
            "You are an AI assistant with access to Salesforce tools. "
            "Use them to help the user with their tasks."
        ),
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get account data by ID",
        user_message="Get information about the account with ID 001gK000003DIn0QAG.",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_account_data_by_id,
                args={
                    "account_id": "001gK000003DIn0QAG",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="account_id", weight=1.0),
        ],
    )

    return suite



import pytest

# Simulates a single block with no children
fake_get_next_page_simple = (
    {
        "results": [
            {
                "object": "block",
                "id": "block1",
                "has_children": False,
                "type": "paragraph",
                "paragraph": {
                    "rich_text": [
                        {
                            "plain_text": "Hello World",
                            "type": "text",
                            "text": {"content": "Hello World", "link": None},
                            "annotations": {
                                "bold": False,
                                "italic": False,
                                "underline": False,
                                "strikethrough": False,
                                "code": False,
                                "color": "default",
                            },
                            "href": None,
                        }
                    ]
                },
            }
        ]
    },
    False,
    None,
)

# Simulates a parent block with a child block
fake_get_next_page_nested = (
    {
        "results": [
            {
                "object": "block",
                "id": "parent_block",
                "has_children": True,
                "type": "paragraph",
                "paragraph": {
                    "rich_text": [
                        {
                            "plain_text": "Parent Block",
                            "type": "text",
                            "text": {"content": "Parent Block", "link": None},
                            "annotations": {
                                "bold": False,
                                "italic": False,
                                "underline": False,
                                "strikethrough": False,
                                "code": False,
                                "color": "default",
                            },
                            "href": None,
                        }
                    ]
                },
            }
        ]
    },
    False,
    None,
)

fake_get_next_page_parent_block = (
    {
        "results": [
            {
                "object": "block",
                "id": "child_block",
                "has_children": False,
                "type": "paragraph",
                "paragraph": {
                    "rich_text": [
                        {
                            "plain_text": "Child Block",
                            "type": "text",
                            "text": {"content": "Child Block", "link": None},
                            "annotations": {
                                "bold": False,
                                "italic": False,
                                "underline": False,
                                "strikethrough": False,
                                "code": False,
                                "color": "default",
                            },
                            "href": None,
                        }
                    ]
                },
            }
        ]
    },
    False,
    None,
)


@pytest.fixture
def setup_notion_pages(monkeypatch):
    from arcade_notion_toolkit.tools import pages

    monkeypatch.setattr(pages, "get_headers", lambda ctx: {"Authorization": "Bearer test"})
    monkeypatch.setattr(
        pages, "get_url", lambda endpoint, block_id=None: f"https://dummy/{block_id}"
    )
    return pages


@pytest.mark.asyncio
async def test_get_page_content_by_id_simple(mock_context, monkeypatch, setup_notion_pages):
    pages = setup_notion_pages

    # Patch get_object_metadata to return a dummy page with title 'Test Page'
    async def fake_get_object_metadata(context, object_id=None, **kwargs):
        return {
            "id": object_id,
            "object": "page",
            "properties": {"title": {"title": [{"plain_text": "Test Page"}]}},
        }

    monkeypatch.setattr(pages, "get_object_metadata", fake_get_object_metadata)

    # Patch get_next_page to
    async def fake_get_next_page(client, url, headers, params, cursor):
        return fake_get_next_page_simple

    monkeypatch.setattr(pages, "get_next_page", fake_get_next_page)

    # Call the function under test
    result = await pages.get_page_content_by_id(mock_context, "test_page_id")
    expected = "# Test Page\nHello World  \n"
    assert result == expected


@pytest.mark.asyncio
async def test_get_page_content_by_id_nested(mock_context, monkeypatch, setup_notion_pages):
    pages = setup_notion_pages

    # Patch get_object_metadata to return a dummy page with title 'Test Nested'
    async def fake_get_object_metadata(context, object_id=None, **kwargs):
        return {
            "id": object_id,
            "object": "page",
            "properties": {"title": {"title": [{"plain_text": "Test Nested"}]}},
        }

    monkeypatch.setattr(pages, "get_object_metadata", fake_get_object_metadata)

    # Patch get_next_page
    async def fake_get_next_page(client, url, headers, params, cursor):
        if url == "https://dummy/test_nested":
            return fake_get_next_page_nested
        elif url == "https://dummy/parent_block":
            return fake_get_next_page_parent_block
        return ({"results": []}, False, None)

    monkeypatch.setattr(pages, "get_next_page", fake_get_next_page)

    # Call the function under test
    result = await pages.get_page_content_by_id(mock_context, "test_nested")
    expected = "# Test Nested\nParent Block  \n    Child Block  \n"
    assert result == expected



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Microsoft

from arcade_outlook_calendar._utils import (
    create_timezone_request_config,
    get_default_calendar_timezone,
    prepare_meeting_body,
    remove_timezone_offset,
    validate_date_times,
    validate_emails,
)
from arcade_outlook_calendar.client import get_client
from arcade_outlook_calendar.models import (
    Attendee,
    DateTimeTimeZone,
    Event,
)


@tool(requires_auth=Microsoft(scopes=["MailboxSettings.Read", "Calendars.ReadWrite"]))
async def create_event(
    context: ToolContext,
    subject: Annotated[str, "The text of the event's subject (title) line."],
    body: Annotated[str, "The body of the event"],
    start_date_time: Annotated[
        str,
        "The datetime of the event's start, represented in "
        "ISO 8601 format. Timezone offset is ignored. For example, 2025-04-25T13:00:00",
    ],
    end_date_time: Annotated[
        str,
        "The datetime of the event's end, represented in "
        "ISO 8601 format. Timezone offset is ignored. For example, 2025-04-25T13:30:00",
    ],
    location: Annotated[str | None, "The location of the event"] = None,
    attendee_emails: Annotated[
        list[str] | None,
        "The email addresses of the attendees of the event. "
        "Must be valid email addresses e.g., username@domain.com.",
    ] = None,
    is_online_meeting: Annotated[
        bool, "Whether the event is an online meeting. Defaults to False"
    ] = False,
    custom_meeting_url: Annotated[
        str | None,
        "The URL of the online meeting. If not provided and is_online_meeting is True, "
        "then a url will be generated for you",
    ] = None,
) -> Annotated[dict, "A dictionary containing the created event details"]:
    """Create an event in the authenticated user's default calendar.

    Ignores timezone offsets provided in the start_date_time and end_date_time parameters.
    Instead, uses the user's default calendar timezone to filter events.
    If the user has not set a timezone for their calendar, then the timezone will be UTC.
    """
    # Validate & cleanse inputs
    validate_emails(attendee_emails or [])
    validate_date_times(start_date_time, end_date_time)
    body, is_online_meeting = prepare_meeting_body(body, custom_meeting_url, is_online_meeting)

    client = get_client(context.get_auth_token_or_empty())

    time_zone = await get_default_calendar_timezone(client)
    start_date_time = remove_timezone_offset(start_date_time)
    end_date_time = remove_timezone_offset(end_date_time)
    event = Event(
        subject=subject,
        body=body,
        start=DateTimeTimeZone(date_time=start_date_time, time_zone=time_zone),
        end=DateTimeTimeZone(date_time=end_date_time, time_zone=time_zone),
        location=location or "",
        attendees=[Attendee(address=attendee) for attendee in attendee_emails or []],
        is_online_meeting=is_online_meeting,
    ).to_sdk()
    request_config = create_timezone_request_config(time_zone)

    response = await client.me.events.post(body=event, request_configuration=request_config)

    return Event.from_sdk(response).to_dict()  # type: ignore[arg-type]



from textwrap import dedent
from typing import Any, Callable, Optional

from crewai.tools.base_tool import BaseTool
from pydantic import BaseModel as PydanticBaseModel


class StructuredTool(BaseTool):  # type: ignore[no-any-unimported]
    """A tool that executes functions with structured inputs using a schema."""

    func: Callable[..., Any]
    """The callable function that implements the tool's functionality."""

    def _run(self, *args: Any, **kwargs: Any) -> Any:
        """Execute the tool's function with the provided arguments."""
        return self.func(*args, **kwargs)

    @classmethod
    def from_function(
        cls,
        func: Callable,
        args_schema: type[PydanticBaseModel],
        name: Optional[str] = None,
        description: Optional[str] = None,
        **kwargs: Any,
    ) -> "StructuredTool":
        """Create a new StructuredTool instance from a function.

        Args:
            func: Function to wrap as a CrewAI tool
            name: Custom name for the tool
            description: Custom description of the tool's functionality
            args_schema: Pydantic model defining the expected argument structure
            **kwargs: Additional tool configuration parameters, like cache_function

        Returns:
            StructuredTool: A new tool instance wrapping the provided function
        """
        name = name or func.__name__

        if description is None:
            description = func.__doc__
            if description is None:
                raise ValueError("Function must have a docstring if description not provided.")
            # Clean up docstring
            description = dedent(description).strip()

        return cls(
            func=func,
            args_schema=args_schema,
            name=name,
            description=description,
            **kwargs,
        )



import json
from unittest.mock import patch

import pytest
from arcade_tdk import ToolContext, ToolSecretItem

from arcade_search.tools import search_google


@pytest.fixture
def mock_context():
    return ToolContext(secrets=[ToolSecretItem(key="serp_api_key", value="fake_api_key")])


@pytest.mark.asyncio
async def test_search_google_success(mock_context):
    with (
        patch("arcade_search.utils.SerpClient") as MockClient,
    ):
        mock_client_instance = MockClient.return_value
        mock_client_instance.search.return_value.as_dict.return_value = {
            "organic_results": [
                {"title": "Result 1", "link": "http://example.com/1"},
                {"title": "Result 2", "link": "http://example.com/2"},
                {"title": "Result 3", "link": "http://example.com/3"},
            ]
        }

        result = await search_google(mock_context, "test query", 2)

        expected_result = json.dumps([
            {"title": "Result 1", "link": "http://example.com/1"},
            {"title": "Result 2", "link": "http://example.com/2"},
        ])
        assert result == expected_result


@pytest.mark.asyncio
async def test_search_google_no_results(mock_context):
    with (
        patch("arcade_search.utils.SerpClient") as MockClient,
    ):
        mock_client_instance = MockClient.return_value
        mock_client_instance.search.return_value.as_dict.return_value = {"organic_results": []}

        result = await search_google(mock_context, "test query", 2)

        expected_result = json.dumps([])
        assert result == expected_result



import asyncio
from typing import Any

from arcade_tdk import ToolContext

from arcade_notion_toolkit.enums import BlockType
from arcade_notion_toolkit.utils import get_page_url


class BlockToMarkdownConverter:
    """
    A converter class that transforms Notion blocks into Markdown.

    The class registers conversion handlers for different Notion block types.
    If a block type does not have a handler, then the block's plain text is returned.
    """

    def __init__(self, context: ToolContext):
        self.context = context
        # block types whose conversion logic has been implemented
        # TODO: implement conversion logic for more block types
        self.handlers = {
            BlockType.BULLETED_LIST_ITEM.value: self._convert_bulleted_list_item,
            BlockType.EQUATION.value: self._convert_equation,
            BlockType.HEADING_1.value: self._convert_heading_1,
            BlockType.HEADING_2.value: self._convert_heading_2,
            BlockType.HEADING_3.value: self._convert_heading_3,
            BlockType.LINK_PREVIEW.value: self._convert_link_preview,
            BlockType.NUMBERED_LIST_ITEM.value: self._convert_numbered_list_item,
            BlockType.PARAGRAPH.value: self._convert_paragraph,
        }

    async def convert_block(self, block: dict[str, Any]) -> str:
        """
        Convert a single Notion block to a Markdown string

        Args:
            block (dict[str, Any]): A Notion block.

        Returns:
            str: A Markdown string.
        """
        block_type = block.get("type")
        if block_type in self.handlers:
            converter = self.handlers[block_type]
            if asyncio.iscoroutinefunction(converter):
                md: str = await converter(block)
                return md
            else:
                return converter(block)
        elif block_type == BlockType.CHILD_PAGE.value:
            return await self._convert_child_page(block)
        else:
            return self._get_plaintext(block)

    @staticmethod
    def rich_text_to_markdown(rich_text_items: list[dict[str, Any]]) -> str:
        """
        Convert a list of rich text items (from a Notion block) into Markdown.

        Handles formatting such as bold, italic, strikethrough, underline (via HTML),
        inline code, text coloring, hyperlinks, and equations.
        """
        md = ""
        for item in rich_text_items:
            annotations = item.get("annotations", {})
            type_val = item.get("type", "text")
            link = None

            # Special handling for inline equations.
            if type_val == "equation":
                expression = item.get("equation", {}).get("expression", "")
                md += f"${expression}$"
                continue

            if type_val == "text":
                text_obj = item.get("text", {})
                text = text_obj.get("content", "")
                link_obj = text_obj.get("link")
                link = (
                    link_obj.get("url")
                    if (link_obj and isinstance(link_obj, dict))
                    else item.get("href")
                )
            elif type_val == "mention":
                text = item.get("plain_text", "")
                link = item.get("href")
            else:
                text = item.get("plain_text", "")
                link = item.get("href")

            if text.strip() == "":
                continue

            # Apply annotation formatting.
            text = BlockToMarkdownConverter.apply_formatting(text, annotations, link)

            md += text

        return md

    @staticmethod
    def apply_formatting(text: str, annotations: dict[str, Any], link: str | None = None) -> str:
        """Apply formatting to a text string based on the annotations.
        Used when converting rich text to markdown

        Args:
            text (str): The text to format.
            annotations (dict[str, Any]): The annotations to apply to the text.
            link (str | None): An optional link for a hyperlink.

        Returns:
            str: The formatted text.
        """
        # If code block, wrap in backticks and skip other formatting.
        if annotations.get("code"):
            return f"`{text}`"

        # Add underline
        if annotations.get("underline"):
            text = f"<u>{text}</u>"

        # Apply color
        color = annotations.get("color", "default")
        if color != "default":
            text = f'<span style="color: {color};">{text}</span>'

        # Add bold, italic, and strikethrough
        markers = [
            marker
            for key, marker in (("bold", "**"), ("italic", "*"), ("strikethrough", "~~"))
            if annotations.get(key)
        ]
        if markers:
            text = "".join(markers) + text + "".join(reversed(markers))

        # Add hyperlink
        if link:
            text = f"[{text}]({link})"

        return text

    def _get_plaintext(self, block: dict[str, Any]) -> str:
        """
        Extract and return the plain text from a Notion block.
        This acts as a fallback for unsupported block types.
        """
        block_type: str = block.get("type", "")
        content = block.get(block_type, {})
        if isinstance(content, dict):
            rich_text_items = content.get("rich_text", [])
            return "".join(item.get("plain_text", "") for item in rich_text_items)
        return ""

    def _convert_text_block(self, block: dict[str, Any], element_key: str, prefix: str = "") -> str:
        """
        Helper method to convert a Notion block's rich_text element into a Markdown string.
        Optionally, a prefix (like a markdown list marker or heading hashes) is added.
        """
        element = block.get(element_key, {})
        rich_text_items = element.get("rich_text", [])
        text = self.rich_text_to_markdown(rich_text_items)
        return f"{prefix}{text}  \n"

    async def _convert_child_page(self, block: dict[str, Any]) -> str:
        """
        Asynchronously convert a child page block. This requires fetching the page's URL.
        """
        page_url = await get_page_url(self.context, block.get("id", ""))
        child_page = block.get("child_page", {})
        rich_text_items = child_page.get("rich_text", [])
        if rich_text_items:
            title = self.rich_text_to_markdown(rich_text_items)
        else:
            title = child_page.get("title", "")
        return f"[{title}]({page_url})  \n"

    def _convert_bulleted_list_item(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "bulleted_list_item", "- ")

    def _convert_equation(self, block: dict[str, Any]) -> str:
        expression = block.get("equation", {}).get("expression", "")
        return f"$$ {expression} $$  \n"

    def _convert_heading_1(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "heading_1", "# ")

    def _convert_heading_2(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "heading_2", "## ")

    def _convert_heading_3(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "heading_3", "### ")

    def _convert_link_preview(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "link_preview")

    def _convert_numbered_list_item(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "numbered_list_item", "1. ")

    def _convert_paragraph(self, block: dict[str, Any]) -> str:
        return self._convert_text_block(block, "paragraph")



from arcade_core.schema import FullyQualifiedName, ToolkitDefinition


def test_initialization():
    fqn = FullyQualifiedName("Tool1", "Toolkit1", "1.0")
    assert fqn.name == "Tool1"
    assert fqn.toolkit_name == "Toolkit1"
    assert fqn.toolkit_version == "1.0"


def test_str():
    fqn = FullyQualifiedName("Tool1", "Toolkit1", "1.0")
    assert str(fqn) == "Toolkit1.Tool1"


def test_equality():
    fqn1 = FullyQualifiedName("Tool1", "Toolkit1", "1.0")
    fqn2 = FullyQualifiedName("Tool1", "Toolkit1", "1.0")
    fqn3 = FullyQualifiedName("Tool2", "Toolkit1", "1.0")
    assert fqn1 == fqn2
    assert fqn1 != fqn3


def test_equality_ignoring_version():
    fqn1 = FullyQualifiedName("Tool1", "Toolkit1", "1.0")
    fqn2 = FullyQualifiedName("Tool1", "Toolkit1", "2.0")
    assert fqn1.equals_ignoring_version(fqn2)


def test_ftqn_case_insensitivity():
    fqn1 = FullyQualifiedName("Tool1", "Toolkit1", "latest")
    fqn2 = FullyQualifiedName("TOOL1", "toolKit1", "LATEST")
    assert fqn1 == fqn2


def test_hash():
    fqn1 = FullyQualifiedName("Tool1", "Toolkit1", "1.0")
    fqn2 = FullyQualifiedName("TOOL1", "toolkit1", "1.0")
    fqn3 = FullyQualifiedName("Tool2", "Toolkit1", "1.0")
    fqn_set = {fqn1, fqn2, fqn3}
    assert len(fqn_set) == 2


def test_from_toolkit():
    toolkit = ToolkitDefinition(name="toolkit1", version="1.0")
    fqn = FullyQualifiedName.from_toolkit("Tool1", toolkit)
    assert fqn.name == "Tool1"
    assert fqn.toolkit_name == "toolkit1"
    assert fqn.toolkit_version == "1.0"



import json
from typing import Annotated

from arcade_core.catalog import ToolCatalog
from arcade_serve.mcp.convert import convert_to_mcp_content, create_mcp_tool
from arcade_tdk import tool


@tool
def sample_tool(x: Annotated[int, "first"], y: Annotated[int, "second"]) -> int:
    """Return x+y"""

    return x + y


def test_convert_to_mcp_content_primitives():
    assert convert_to_mcp_content(42) == [{"type": "text", "text": "42"}]
    assert convert_to_mcp_content("hello") == [{"type": "text", "text": "hello"}]
    assert convert_to_mcp_content(True) == [{"type": "text", "text": "True"}]


def test_convert_to_mcp_content_complex():
    data = {"a": 1}
    expected_json = json.dumps(data)
    assert convert_to_mcp_content(data) == [{"type": "text", "text": expected_json}]


def test_create_mcp_tool():
    # Materialize a tool via catalog then feed it to create_mcp_tool
    catalog = ToolCatalog()
    catalog.add_tool(sample_tool, "convert_toolkit")
    mat_tool = next(iter(catalog))  # only tool
    mcp_tool = create_mcp_tool(mat_tool)

    assert mcp_tool is not None
    assert mcp_tool["name"] == "ConvertToolkit_SampleTool"
    assert mcp_tool["description"]
    # Ensure input schema contains both parameters and marks them required
    props = mcp_tool["inputSchema"]["properties"]
    assert set(props.keys()) == {"x", "y"}

    required_fields = set(mcp_tool["inputSchema"].get("required", []))
    # Ensure no unexpected required fields and that declared ones are subset of expected
    assert required_fields.issubset({"x", "y"})



from collections.abc import Callable

import pytest
from arcade_tdk import ToolContext

from arcade_jira.exceptions import MultipleItemsFoundError, NotFoundError
from arcade_jira.utils import clean_project_dict, find_unique_project


@pytest.mark.asyncio
async def test_find_unique_project_by_id_success(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    fake_cloud_name: str,
):
    sample_project = build_project_dict()
    project_response = mock_httpx_response(200, sample_project)
    mock_httpx_client.get.return_value = project_response

    response = await find_unique_project(mock_context, sample_project["id"])
    assert response == clean_project_dict(sample_project, fake_cloud_name)


@pytest.mark.asyncio
async def test_find_unique_project_by_name_with_a_single_match(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_project_search_response_dict: Callable,
    fake_cloud_name: str,
):
    sample_project = build_project_dict()
    get_project_by_id_response = mock_httpx_response(404, {})
    get_projects_without_id_response = mock_httpx_response(
        200, build_project_search_response_dict([sample_project])
    )
    mock_httpx_client.get.side_effect = [
        get_project_by_id_response,
        get_projects_without_id_response,
    ]

    response = await find_unique_project(mock_context, sample_project["name"].lower())
    assert response == clean_project_dict(sample_project, fake_cloud_name)


@pytest.mark.asyncio
async def test_find_unique_project_by_name_with_multiple_matches(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    build_project_dict: Callable,
    build_project_search_response_dict: Callable,
    generate_random_str: Callable,
):
    project_name = generate_random_str()
    sample_projects = [
        build_project_dict(name=project_name),
        build_project_dict(name=project_name),
    ]
    get_project_by_id_response = mock_httpx_response(404, {})
    search_projects_response = mock_httpx_response(
        200, build_project_search_response_dict(sample_projects)
    )
    mock_httpx_client.get.side_effect = [
        get_project_by_id_response,
        search_projects_response,
    ]

    with pytest.raises(MultipleItemsFoundError) as exc:
        await find_unique_project(mock_context, sample_projects[0]["name"].lower())

    assert sample_projects[0]["id"] in exc.value.message
    assert sample_projects[1]["id"] in exc.value.message


@pytest.mark.asyncio
async def test_find_unique_project_by_name_without_matches(
    mock_context: ToolContext,
    mock_httpx_client,
    mock_httpx_response: Callable,
    generate_random_str: Callable,
    build_project_search_response_dict: Callable,
):
    get_project_by_id_response = mock_httpx_response(404, {})
    search_projects_response = mock_httpx_response(200, build_project_search_response_dict([]))
    mock_httpx_client.get.side_effect = [
        get_project_by_id_response,
        search_projects_response,
    ]

    with pytest.raises(NotFoundError):
        await find_unique_project(mock_context, generate_random_str())



import arcade_google_docs.doc_to_html as doc_to_html


def convert_document_to_markdown(document: dict) -> str:
    md = f"---\ntitle: {document['title']}\ndocumentId: {document['documentId']}\n---\n"
    for element in document["body"]["content"]:
        md += convert_structural_element(element)
    return md


def convert_structural_element(element: dict) -> str:
    if "sectionBreak" in element or "tableOfContents" in element:
        return ""

    elif "paragraph" in element:
        md = ""
        prepend = get_paragraph_style_prepend_str(element["paragraph"]["paragraphStyle"])
        for item in element["paragraph"]["elements"]:
            if "textRun" not in item:
                continue
            content = extract_paragraph_content(item["textRun"])
            md += f"{prepend}{content}"
        return md

    elif "table" in element:
        return doc_to_html.convert_structural_element(element)

    else:
        raise ValueError(f"Unknown document body element type: {element}")


def extract_paragraph_content(text_run: dict) -> str:
    content = text_run["content"]
    style = text_run["textStyle"]
    return apply_text_style(content, style)


def apply_text_style(content: str, style: dict) -> str:
    append = "\n" if content.endswith("\n") else ""
    content = content.rstrip("\n")
    italic = style.get("italic", False)
    bold = style.get("bold", False)
    if italic:
        content = f"_{content}_"
    if bold:
        content = f"**{content}**"
    return f"{content}{append}"


def get_paragraph_style_prepend_str(style: dict) -> str:
    named_style = style["namedStyleType"]
    if named_style == "NORMAL_TEXT":
        return ""
    elif named_style == "TITLE":
        return "# "
    elif named_style == "SUBTITLE":
        return "## "
    elif named_style.startswith("HEADING_"):
        try:
            heading_level = int(named_style.split("_")[1])
            return f"{'#' * heading_level} "
        except ValueError:
            return ""
    return ""



from typing import NewType


class PositiveNonZeroInt(int):
    def __new__(cls, value: str | int, name: str = "value") -> "PositiveNonZeroInt":
        def validate(val: int) -> int:
            if val < 1:
                raise ValueError(f"{name} must be a positive non-zero integer, got {val}")
            return val

        try:
            value = int(value)
        except ValueError:
            raise ValueError(f"{name} must be a valid integer, got {value!r}")

        validated_value = validate(value)
        instance = super().__new__(cls, validated_value)
        return instance


SlackOffsetSecondsFromUTC = NewType("SlackOffsetSecondsFromUTC", int)  # observe it can be negative
SlackPaginationNextCursor = str | None
SlackUserFieldId = NewType("SlackUserFieldId", str)
SlackUserId = NewType("SlackUserId", str)
SlackTeamId = NewType("SlackTeamId", str)
SlackTimestampStr = NewType("SlackTimestampStr", str)



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_dropbox
from arcade_dropbox.constants import ItemCategory
from arcade_dropbox.critics import DropboxPathCritic
from arcade_dropbox.tools.browse import search_files_and_folders

rubric = EvalRubric(
    fail_threshold=0.8,
    warn_threshold=0.9,
)


catalog = ToolCatalog()
catalog.add_module(arcade_dropbox)


@tool_eval()
def search_files_and_folders_eval_suite() -> EvalSuite:
    """Create an evaluation suite for the search_files_and_folders tool."""
    suite = EvalSuite(
        name="list_items_in_folder",
        system_message="You are an AI assistant that can interact with files and folders in Dropbox using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Search for files about 'quarterly report' in my Dropbox",
        user_message="Search for files about 'quarterly report' in my Dropbox",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_files_and_folders,
                args={
                    "keywords": "quarterly report",
                    "search_in_folder_path": None,
                    "filter_by_category": None,
                    "limit": 100,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.6),
            BinaryCritic(critic_field="search_in_folder_path", weight=0.1),
            BinaryCritic(critic_field="filter_by_category", weight=0.1),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="cursor", weight=0.1),
        ],
    )

    suite.add_case(
        name="Search for files about 'quarterly report' in a sub-folder",
        user_message="Search for files about 'quarterly report' in the folder AcmeInc/Reports",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_files_and_folders,
                args={
                    "keywords": "quarterly report",
                    "search_in_folder_path": "/AcmeInc/Reports",
                    "filter_by_category": None,
                    "limit": 100,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.35),
            DropboxPathCritic(critic_field="search_in_folder_path", weight=0.35),
            BinaryCritic(critic_field="filter_by_category", weight=0.1),
            BinaryCritic(critic_field="limit", weight=0.1),
            BinaryCritic(critic_field="cursor", weight=0.1),
        ],
    )

    suite.add_case(
        name="Search for PDF files about 'quarterly report' in a sub-folder",
        user_message="Search for PDF files about 'quarterly report' in the folder AcmeInc/Reports",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_files_and_folders,
                args={
                    "keywords": "quarterly report",
                    "search_in_folder_path": "/AcmeInc/Reports",
                    "filter_by_category": [ItemCategory.PDF.value],
                    "limit": 100,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.25),
            DropboxPathCritic(critic_field="search_in_folder_path", weight=0.25),
            BinaryCritic(critic_field="filter_by_category", weight=0.25),
            BinaryCritic(critic_field="limit", weight=0.125),
            BinaryCritic(critic_field="cursor", weight=0.125),
        ],
    )

    suite.add_case(
        name="Search for PDF files about 'quarterly report' in a sub-folder",
        user_message="Return the first 10 PDF files about 'quarterly report' in the folder AcmeInc/Reports",
        expected_tool_calls=[
            ExpectedToolCall(
                func=search_files_and_folders,
                args={
                    "keywords": "quarterly report",
                    "search_in_folder_path": "/AcmeInc/Reports",
                    "filter_by_category": [ItemCategory.PDF.value],
                    "limit": 10,
                    "cursor": None,
                },
            ),
        ],
        critics=[
            BinaryCritic(critic_field="keywords", weight=0.2),
            DropboxPathCritic(critic_field="search_in_folder_path", weight=0.2),
            BinaryCritic(critic_field="filter_by_category", weight=0.2),
            BinaryCritic(critic_field="limit", weight=0.2),
            BinaryCritic(critic_field="cursor", weight=0.2),
        ],
    )

    return suite



from typing import Annotated, Optional, Union

import pytest
from arcade_core.catalog import ToolCatalog
from arcade_core.schema import (
    InputParameter,
    ToolInput,
    ToolOutput,
    ValueSchema,
)
from arcade_tdk import tool
from pydantic import BaseModel, Field


class ProductOutput(BaseModel):
    product_name: str = Field(..., description="The name of the product")
    price: int = Field(..., description="The price of the product")
    stock_quantity: int = Field(..., description="The stock quantity of the product")


@tool(desc="A function that returns a Pydantic model")
def func_returns_pydantic_model() -> Annotated[ProductOutput, "The product, price, and quantity"]:
    return ProductOutput(
        product_name="Product 1",
        price=100,
        stock_quantity=1000,
    )


@tool(desc="A function that accepts a required Pydantic Field with a description")
def func_takes_pydantic_field_with_description(
    product_name: str = Field(..., description="The name of the product"),
) -> str:
    return product_name


@tool(desc="A function that accepts an optional Pydantic Field")
def func_takes_pydantic_field_optional(
    product_name: Optional[str] = Field(None, description="The name of the product"),
) -> str:
    return product_name


@tool(desc="A function that accepts an optional Pydantic Field with bar syntax")
def func_takes_pydantic_field_optional_bar_syntax(
    product_name: str | None = Field(None, description="The name of the product"),
) -> str:
    return product_name


@tool(desc="A function that accepts an optional Pydantic Field with union syntax")
def func_takes_pydantic_field_optional_union_syntax(
    product_name: Union[str, None] = Field(None, description="The name of the product"),
) -> str:
    return product_name


# Annotated[] takes precedence over Field() properties
@tool(desc="A function that accepts an annotated Pydantic Field")
def func_takes_pydantic_field_annotated_description(
    product_name: Annotated[str, "The name of the product"] = Field(
        ..., description="The name of the product???"
    ),
) -> str:
    return product_name


# Annotated[] takes precedence over Field() properties
@tool(desc="A function that accepts an annotated Pydantic Field")
def func_takes_pydantic_field_annotated_name_and_description(
    product_name: Annotated[str, "ProductName", "The name of the product"] = Field(
        ..., title="The name of the product???"
    ),
) -> str:
    return product_name


@tool(desc="A function that accepts a Pydantic Field with a default value")
def func_takes_pydantic_field_default(
    product_name: str = Field(description="The name of the product", default="Product 1"),
) -> str:
    return product_name


@tool(desc="A function that accepts a Pydantic Field with a default value factory")
def func_takes_pydantic_field_default_factory(
    product_name: str = Field(
        ..., description="The name of the product", default_factory=lambda: "Product 1"
    ),
) -> str:
    return product_name


# TODO: Function that takes a Pydantic model as an argument: break it down into components? Look at OpenAPI, do they represent nested arguments?
# TODO: Should title and default_value be added to JSON schema?
# TODO: Pydantic Field() properties stretch goal: gt, ge, lt, le, multiple_of, range, regex, max_length, min_length, max_items, min_items, unique_items, exclusive_maximum, exclusive_minimum, title?


### A complex, real-world example
class ProductFilter(BaseModel):
    column: str = Field(..., description="The column to filter on")


class FilterRating(ProductFilter):
    greater_than: int = Field(..., description="The rating to filter greater than", gt=0, lt=5)


class FilterPriceGreaterThan(ProductFilter):
    price: int = Field(..., description="The price to filter greater than", gt=0)


class FilterPriceLessThan(ProductFilter):
    price: int = Field(..., description="The price to filter less than", gt=0)


class ProductSearch(BaseModel):
    column: str = Field("Product Name", description="The column to search in")
    query: str = Field(..., description="The query to search for")
    filter_operation: Union[FilterRating, FilterPriceGreaterThan, FilterPriceLessThan] = None


class ProductOutput(BaseModel):
    product_name: str = Field(..., description="The name of the product")
    price: int = Field(..., description="The price of the product")
    stock_quantity: int = Field(..., description="The stock quantity of the product")


@tool
def read_products(
    action: Annotated[ProductSearch, "The search query to perform"],
    cols: list[str] = Field(
        ...,
        description="The columns to return",
        default_factory=lambda: ["Product Name", "Price", "Stock Quantity"],
    ),
) -> Annotated[list[ProductOutput], "Data with the selected columns"]:
    """Used to search through products by name and filter by rating or price."""

    pass


@pytest.mark.parametrize(
    "func_under_test, expected_tool_def_fields",
    [
        pytest.param(
            func_returns_pydantic_model,
            {
                "output": ToolOutput(
                    value_schema=ValueSchema(val_type="json", enum=None),
                    available_modes=["value", "error"],
                    description="The product, price, and quantity",
                )
            },
            id="func_returns_pydantic_model",
        ),
        pytest.param(
            func_takes_pydantic_field_with_description,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",
                            required=True,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                )
            },
            id="func_takes_pydantic_field_with_description",
        ),
        pytest.param(
            func_takes_pydantic_field_optional,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",
                            required=False,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                )
            },
            id="func_takes_pydantic_field_optional",
        ),
        pytest.param(
            func_takes_pydantic_field_optional_bar_syntax,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",
                            required=False,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                )
            },
            id="func_takes_pydantic_field_optional_bar_syntax",
        ),
        pytest.param(
            func_takes_pydantic_field_optional_union_syntax,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",
                            required=False,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                )
            },
            id="func_takes_pydantic_field_optional_union_syntax",
        ),
        pytest.param(
            func_takes_pydantic_field_annotated_description,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",  # Annotated[] takes precedence over Field() properties
                            required=True,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                )
            },
            id="func_takes_pydantic_field_annotated_description",
        ),
        pytest.param(
            func_takes_pydantic_field_annotated_name_and_description,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="ProductName",
                            description="The name of the product",  # Annotated[] takes precedence over Field() properties
                            required=True,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                )
            },
            id="func_takes_pydantic_field_annotated_name_and_description",
        ),
        pytest.param(
            func_takes_pydantic_field_default,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",
                            required=False,  # Because it has a default value
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                ),
            },
            id="func_takes_pydantic_field_default",
        ),
        pytest.param(
            func_takes_pydantic_field_default_factory,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="product_name",
                            description="The name of the product",
                            required=False,  # Because it has a default value factory
                            inferrable=True,
                            value_schema=ValueSchema(val_type="string", enum=None),
                        )
                    ]
                ),
            },
            id="func_takes_pydantic_field_default_factory",
        ),
        pytest.param(
            read_products,
            {
                "input": ToolInput(
                    parameters=[
                        InputParameter(
                            name="action",
                            description="The search query to perform",
                            required=True,
                            inferrable=True,
                            value_schema=ValueSchema(val_type="json", enum=None),
                        ),
                        InputParameter(
                            name="cols",
                            description="The columns to return",
                            required=False,
                            value_schema=ValueSchema(
                                val_type="array", inner_val_type="string", enum=None
                            ),
                        ),
                    ]
                ),
                "output": ToolOutput(
                    value_schema=ValueSchema(val_type="array", inner_val_type="json", enum=None),
                    available_modes=["value", "error"],
                    description="Data with the selected columns",
                ),
            },
            id="read_products",
        ),
    ],
)
def test_create_tool_def_from_pydantic(func_under_test, expected_tool_def_fields):
    tool_def = ToolCatalog.create_tool_definition(func_under_test, "1.0")

    for field, expected_value in expected_tool_def_fields.items():
        assert getattr(tool_def, field) == expected_value



from unittest.mock import patch

import pytest
from arcade_tdk.errors import ToolExecutionError

from arcade_confluence.client import ConfluenceClientV1


@pytest.fixture
def client_v1() -> ConfluenceClientV1:
    """Fixture that provides a ConfluenceClientV1 instance with mocked cloud_id."""
    with patch("arcade_confluence.client.ConfluenceClient._get_cloud_id", return_value=None):
        return ConfluenceClientV1("fake-token")


@pytest.mark.parametrize(
    "query, enable_fuzzy, expected_cql",
    [
        ("foo", False, '(text ~ "foo" OR title ~ "foo" OR space.title ~ "foo")'),
        ("foo bar", False, '(text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar")'),
        ("foo", True, '(text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~")'),
        ("foo bar", True, '(text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar")'),
    ],
)
def test_build_query_cql(client_v1: ConfluenceClientV1, query, enable_fuzzy, expected_cql) -> None:
    cql = client_v1._build_query_cql(query, enable_fuzzy)
    assert cql == expected_cql


@pytest.mark.parametrize(
    "queries, enable_fuzzy, expected_cql",
    [
        (
            ["foo", "foo bar"],
            False,
            '((text ~ "foo" OR title ~ "foo" OR space.title ~ "foo") AND (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
        (
            ["foo", "foo bar"],
            True,
            '((text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~") AND (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
    ],
)
def test_build_and_cql(client_v1: ConfluenceClientV1, queries, enable_fuzzy, expected_cql) -> None:
    cql = client_v1._build_and_cql(queries, enable_fuzzy)
    assert cql == expected_cql


@pytest.mark.parametrize(
    "queries, enable_fuzzy, expected_cql",
    [
        (
            ["foo", "foo bar"],
            False,
            '((text ~ "foo" OR title ~ "foo" OR space.title ~ "foo") OR (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
        (
            ["foo", "foo bar"],
            True,
            '((text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~") OR (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
    ],
)
def test_build_or_cql(client_v1: ConfluenceClientV1, queries, enable_fuzzy, expected_cql) -> None:
    cql = client_v1._build_or_cql(queries, enable_fuzzy)
    assert cql == expected_cql


@pytest.mark.parametrize(
    "must_contain_all, can_contain_any, enable_fuzzy, expected_cql",
    [
        (None, None, False, ""),
        (
            ["foo", "foo bar"],
            None,
            False,
            '((text ~ "foo" OR title ~ "foo" OR space.title ~ "foo") AND (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
        (
            ["foo", "foo bar"],
            None,
            True,
            '((text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~") AND (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
        (
            None,
            ["foo", "foo bar"],
            False,
            '((text ~ "foo" OR title ~ "foo" OR space.title ~ "foo") OR (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
        (
            None,
            ["foo", "foo bar"],
            True,
            '((text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~") OR (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar"))',  # noqa: E501
        ),
        (
            ["foo", "foo bar"],
            ["foo", "foo bar"],
            False,
            '(((text ~ "foo" OR title ~ "foo" OR space.title ~ "foo") AND (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar")) AND ((text ~ "foo" OR title ~ "foo" OR space.title ~ "foo") OR (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar")))',  # noqa: E501
        ),
        (
            ["foo", "foo bar"],
            ["foo", "foo bar"],
            True,
            '(((text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~") AND (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar")) AND ((text ~ "foo~" OR title ~ "foo~" OR space.title ~ "foo~") OR (text ~ "foo bar" OR title ~ "foo bar" OR space.title ~ "foo bar")))',  # noqa: E501
        ),
    ],
)
def test_construct_cql(
    client_v1: ConfluenceClientV1, must_contain_all, can_contain_any, enable_fuzzy, expected_cql
) -> None:
    if not expected_cql:
        with pytest.raises(ToolExecutionError):
            client_v1.construct_cql(must_contain_all, can_contain_any, enable_fuzzy)
    else:
        cql = client_v1.construct_cql(must_contain_all, can_contain_any, enable_fuzzy)
        assert cql == expected_cql



from enum import Enum
from typing import Any
from urllib.parse import parse_qs, urlparse

import httpx
from arcade_tdk.errors import ToolExecutionError

from arcade_confluence.enums import BodyFormat, PageUpdateMode
from arcade_confluence.utils import (
    build_child_url,
    build_hierarchy,
    remove_none_values,
)


class ConfluenceAPIVersion(str, Enum):
    V1 = "wiki/rest/api"
    V2 = "wiki/api/v2"


class ConfluenceClient:
    ACCESSIBLE_RESOURCES_URL = "https://api.atlassian.com/oauth/token/accessible-resources"
    BASE_URL = "https://api.atlassian.com/ex/confluence"

    def __init__(self, token: str, api_version: ConfluenceAPIVersion):
        self.token = token
        self.cloud_id = self._get_cloud_id()
        self.api_version = api_version.value

    def _get_cloud_id(self) -> str:
        """
        Fetch the cloudId for <workspace_name>.atlassian.net
        using the OAuth2 3LO accessible-resources endpoint.

        For details on why this is necessary, see: https://developer.atlassian.com/cloud/oauth/getting-started/making-calls-to-api
        """
        headers = {"Authorization": f"Bearer {self.token}"}
        resp = httpx.get(self.ACCESSIBLE_RESOURCES_URL, headers=headers)
        resp.raise_for_status()
        resp_json = resp.json()

        if len(resp_json) == 0:
            raise ToolExecutionError(message="No workspaces found for the authenticated user.")

        return resp_json[0].get("id")  # type: ignore[no-any-return]

    async def request(self, method: str, path: str, **kwargs: Any) -> Any:
        headers = {
            "Accept": "application/json",
            "Authorization": f"Bearer {self.token}",
        }
        async with httpx.AsyncClient() as client:
            response = await client.request(
                method,
                f"{self.BASE_URL}/{self.cloud_id}/{self.api_version}/{path.lstrip('/')}",
                headers=headers,
                **kwargs,
            )
            response.raise_for_status()
            return response.json()

    async def get(self, path: str, **kwargs: Any) -> Any:
        return await self.request("GET", path, **kwargs)

    async def post(self, path: str, **kwargs: Any) -> Any:
        return await self.request("POST", path, **kwargs)

    async def put(self, path: str, **kwargs: Any) -> Any:
        return await self.request("PUT", path, **kwargs)


class ConfluenceClientV1(ConfluenceClient):
    def __init__(self, token: str):
        super().__init__(token, api_version=ConfluenceAPIVersion.V1)

    def _build_query_cql(self, query: str, enable_fuzzy: bool) -> str:
        """Build CQL for a single query (term or phrase).

        Args:
            query: The search query (single word term or multi-word phrase)
            enable_fuzzy: Whether to enable fuzzy matching for single terms

        Returns:
            CQL string for the query
        """
        query = query.strip()
        if not query:
            return ""

        # For phrases (multiple words), don't use fuzzy matching
        if " " in query:
            return f'(text ~ "{query}" OR title ~ "{query}" OR space.title ~ "{query}")'
        else:
            # For single terms, optionally use fuzzy matching
            term_suffix = "~" if enable_fuzzy else ""
            return f'(text ~ "{query}{term_suffix}" OR title ~ "{query}{term_suffix}" OR space.title ~ "{query}{term_suffix}")'  # noqa: E501

    def _build_and_cql(self, queries: list[str], enable_fuzzy: bool) -> str:
        """Build CQL for queries that must ALL be present (AND logic).

        Args:
            queries: List of queries that must all be present
            enable_fuzzy: Whether to enable fuzzy matching for single terms

        Returns:
            CQL string with AND logic
        """
        and_parts = []
        for query in queries:
            query_cql = self._build_query_cql(query, enable_fuzzy)
            if query_cql:
                and_parts.append(query_cql)

        if not and_parts:
            return ""

        return f"({' AND '.join(and_parts)})"

    def _build_or_cql(self, queries: list[str], enable_fuzzy: bool) -> str:
        """Build CQL for queries where ANY can be present (OR logic).

        Args:
            queries: List of queries where any can be present
            enable_fuzzy: Whether to enable fuzzy matching for single terms

        Returns:
            CQL string with OR logic
        """
        or_parts = []
        for query in queries:
            query_cql = self._build_query_cql(query, enable_fuzzy)
            if query_cql:
                or_parts.append(query_cql)

        if not or_parts:
            return ""

        return f"({' OR '.join(or_parts)})"

    def construct_cql(
        self,
        must_contain_all: list[str] | None,
        can_contain_any: list[str] | None,
        enable_fuzzy: bool = False,
    ) -> str:
        """Construct CQL query with AND/OR logic.

        Learn about advanced searching using CQL here: https://developer.atlassian.com/cloud/confluence/advanced-searching-using-cql/

        Args:
            must_contain_all: Queries that must ALL be present (AND logic)
            can_contain_any: Queries where ANY can be present (OR logic)
            enable_fuzzy: Whether to enable fuzzy matching for single terms

        Returns:
            CQL query string

        Raises:
            ToolExecutionError: If no search parameters are provided
        """
        cql_parts = []

        # Handle must_contain_all (AND logic)
        if must_contain_all:
            and_cql = self._build_and_cql(must_contain_all, enable_fuzzy)
            if and_cql:
                cql_parts.append(and_cql)

        # Handle can_contain_any (OR logic)
        if can_contain_any:
            or_cql = self._build_or_cql(can_contain_any, enable_fuzzy)
            if or_cql:
                cql_parts.append(or_cql)

        # If there's only one part, return it
        if len(cql_parts) == 1:
            return cql_parts[0]

        # AND the must_contain_all with the can_contain_any
        if len(cql_parts) > 1:
            return f"({' AND '.join(cql_parts)})"

        raise ToolExecutionError(message="At least one search parameter must be provided")

    def transform_search_content_response(
        self, response: dict[str, Any]
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Transform the response from the GET /search endpoint by converting relative webui paths
        to absolute URLs using the base URL from the response.
        """
        base_url = response.get("_links", {}).get("base", "")
        transformed_results = []
        for result in response.get("results", []):
            content = result.get("content", {})
            transformed_result = {
                "id": content.get("id"),
                "title": content.get("title"),
                "type": content.get("type"),
                "status": content.get("status"),
                "excerpt": result.get("excerpt"),
                "url": f"{base_url}{result.get('url')}",
            }
            transformed_results.append(transformed_result)

        return {"results": transformed_results}


class ConfluenceClientV2(ConfluenceClient):
    def __init__(self, token: str):
        super().__init__(token, api_version=ConfluenceAPIVersion.V2)

    def _transform_links(
        self, response: dict[str, Any], base_url: str | None = None
    ) -> dict[str, Any]:
        """
        Transform the links in a page response by converting relative URLs to absolute URLs.

        Args:
            response: A page object from the API
            base_url: The base URL to use for the transformation

        Returns:
            The transformed response
        """
        result = response.copy()
        if "_links" in result:
            base_url = base_url or result["_links"].get("base", "")
            webui_path = result["_links"].get("webui", "")
            result["url"] = f"{base_url}{webui_path}"
            del result["_links"]
        return result

    def transform_get_spaces_response(
        self, response: dict[str, Any]
    ) -> dict[str, list[dict[str, Any]]]:
        """
        Transform the response from the GET /spaces endpoint by converting relative webui paths
        to absolute URLs using the base URL from the response.
        """
        pagination_token = parse_qs(urlparse(response.get("_links", {}).get("next", "")).query).get(
            "cursor",
            [None],  # type: ignore[list-item]
        )[0]

        base_url = response.get("_links", {}).get("base", "")
        results = response.get("results", [])

        transformed_results = []
        for space in results:
            space_copy = space.copy()
            if "_links" in space_copy and "webui" in space_copy["_links"]:
                webui_path = space_copy["_links"]["webui"]
                space_copy["url"] = base_url + webui_path
                del space_copy["_links"]
            transformed_results.append(space_copy)

        results = {"spaces": transformed_results, "pagination_token": pagination_token}
        return remove_none_values(results)

    def transform_list_pages_response(self, response: dict[str, Any]) -> dict[str, Any]:
        """Transform the response from the GET /pages endpoint."""
        pagination_token = parse_qs(urlparse(response.get("_links", {}).get("next", "")).query).get(
            "cursor",
            [None],  # type: ignore[list-item]
        )[0]

        base_url = response.get("_links", {}).get("base", "")
        pages = [self._transform_links(page, base_url) for page in response["results"]]
        results = {"pages": pages, "pagination_token": pagination_token}
        return remove_none_values(results)

    def transform_get_multiple_pages_response(
        self, response: dict[str, Any]
    ) -> dict[str, list[dict[str, Any]]]:
        """Transform the response from the GET /pages endpoint."""
        base_url = response.get("_links", {}).get("base", "")
        pages = [self._transform_links(page, base_url) for page in response["results"]]
        return {"pages": pages}

    def transform_space_response(
        self, response: dict[str, Any], base_url: str | None = None
    ) -> dict[str, dict[str, Any]]:
        """Transform API responses that return a space object."""
        return {"space": self._transform_links(response, base_url)}

    def transform_page_response(self, response: dict[str, Any]) -> dict[str, dict[str, Any]]:
        """Transform API responses that return a page object."""
        return {"page": self._transform_links(response)}

    def transform_get_attachments_response(self, response: dict[str, Any]) -> dict[str, Any]:
        """Transform the response from the GET /pages/{id}/attachments endpoint."""
        pagination_token = parse_qs(urlparse(response.get("_links", {}).get("next", "")).query).get(
            "cursor",
            [None],  # type: ignore[list-item]
        )[0]

        base_url = response.get("_links", {}).get("base", "")
        attachments = []
        for attachment in response["results"]:
            result = attachment.copy()
            if "_links" in result:
                webui_path = result["_links"].get("webui", "")
                download_path = result["_links"].get("download", "")
                result["url"] = f"{base_url}{webui_path}"
                result["download_link"] = f"{base_url}{download_path}"
                del result["_links"]
                del result["webuiLink"]
                del result["downloadLink"]
                del result["version"]
            attachments.append(result)

        return {"attachments": attachments, "pagination_token": pagination_token}

    def prepare_update_page_payload(
        self,
        page_id: str,
        status: str,
        title: str,
        body_representation: str,
        body_value: str,
        version_number: int,
        version_message: str,
    ) -> dict[str, Any]:
        """Prepare a payload for the PUT /pages/{id} endpoint."""
        return {
            "id": page_id,
            "status": status,
            "title": title,
            "body": {
                "representation": body_representation,
                "value": body_value,
            },
            "version": {
                "number": version_number,
                "message": version_message,
            },
        }

    def prepare_update_page_content_payload(
        self,
        content: str,
        update_mode: PageUpdateMode,
        old_content: str,
        page_id: str,
        status: str,
        title: str,
        body_representation: BodyFormat,
        old_version_number: int,
    ) -> dict[str, Any]:
        """Prepare a payload for when updating the content of a page

        Args:
            content: The content to update the page with
            update_mode: The mode of update to use
            old_content: The content of the page before the update
            page_id: The ID of the page to update
            status: The status of the page
            title: The title of the page
            body_representation: The format that the body (content) is in
            old_version_number: The version number of the page before the update

        Returns:
            A payload for the PUT /pages/{id} endpoint's json body
        """
        updated_content = ""
        updated_message = ""
        if update_mode == PageUpdateMode.APPEND:
            updated_content = f"{old_content}<br/>{content}"
            updated_message = "Append content to the page"
        elif update_mode == PageUpdateMode.PREPEND:
            updated_content = f"{content}<br/>{old_content}"
            updated_message = "Prepend content to the page"
        elif update_mode == PageUpdateMode.REPLACE:
            updated_content = content
            updated_message = "Replace the page content"
        payload = self.prepare_update_page_payload(
            page_id=page_id,
            status=status,
            title=title,
            body_representation=body_representation.to_api_value(),
            body_value=updated_content,
            version_number=old_version_number + 1,
            version_message=updated_message,
        )
        return payload

    async def get_root_pages_in_space(self, space_id: str) -> dict[str, Any]:
        """
        Get the root pages in a space.

        Requires Confluence scope 'read:page:confluence'
        """
        params = {
            "depth": "root",
            "limit": 250,
        }
        pages = await self.get(f"spaces/{space_id}/pages", params=params)
        base_url = pages.get("_links", {}).get("base", "")
        return {"pages": [self._transform_links(page, base_url) for page in pages["results"]]}

    async def get_space_homepage(self, space_id: str) -> dict[str, Any]:
        """
        Get the homepage of a space.

        Requires Confluence scope 'read:page:confluence'
        """
        root_pages = await self.get_root_pages_in_space(space_id)
        for page in root_pages["pages"]:
            if page.get("url", "").endswith("overview"):
                return self._transform_links(page)
        raise ToolExecutionError(message="No homepage found for space.")

    async def get_page_by_id(
        self, page_id: str, content_format: BodyFormat = BodyFormat.STORAGE
    ) -> dict[str, Any]:
        """Get a page by its ID.

        Requires Confluence scope 'read:page:confluence'

        Args:
            page_id: The ID of the page to get
            content_format: The format of the page content

        Returns:
            The page object
        """
        params = remove_none_values({
            "body-format": content_format.to_api_value(),
        })
        try:
            page = await self.get(f"pages/{page_id}", params=params)
        except httpx.HTTPStatusError as e:
            # If the page is not found, return an empty page object
            if e.response.status_code in [400, 404]:
                return self.transform_page_response({})
            raise

        return self.transform_page_response(page)

    async def get_page_by_title(
        self, page_title: str, content_format: BodyFormat = BodyFormat.STORAGE
    ) -> dict[str, Any]:
        """Get a page by its title.

        Requires Confluence scope 'read:page:confluence'

        Args:
            page_title: The title of the page to get
            content_format: The format of the page content

        Returns:
            The page object
        """
        params = {
            "title": page_title,
            "body-format": content_format.to_api_value(),
        }
        response = await self.get("pages", params=params)
        pages = response.get("results", [])
        if not pages:
            # If the page is not found, return an empty page object
            return self.transform_page_response({})
        return self.transform_page_response(pages[0])

    async def get_space_by_id(self, space_id: str) -> dict[str, Any]:
        """Get a space by its ID.

        Requires Confluence scope 'read:space:confluence'

        Args:
            space_id: The ID of the space to get

        Returns:
            The space object
        """
        space = await self.get(f"spaces/{space_id}")
        return self.transform_space_response(space)

    async def get_space_by_key(self, space_key: str) -> dict[str, Any]:
        """Get a space by its key.

        Requires Confluence scope 'read:space:confluence'

        Args:
            space_key: The key of the space to get

        Returns:
            The space object
        """
        response = await self.get("spaces", params={"keys": [space_key]})
        base_url = response.get("_links", {}).get("base", "")
        spaces = response.get("results", [])
        if not spaces:
            raise ToolExecutionError(message=f"No space found with key: '{space_key}'")
        return self.transform_space_response(spaces[0], base_url=base_url)

    async def get_space(self, space_identifier: str) -> dict[str, Any]:
        """Get a space from its identifier.

        Requires Confluence scope 'read:space:confluence'

        Args:
            space_identifier: The identifier of the space to get. Can be a space's ID or key.
        """
        return (
            await self.get_space_by_id(space_identifier)
            if space_identifier.isdigit()
            else await self.get_space_by_key(space_identifier)
        )

    async def get_page_id(self, page_identifier: str) -> str:
        """Get the ID of a page from its identifier.

        Args:
            page_identifier: The identifier of the page to get. Can be a page's ID or title.

        Returns:
            The ID of the page
        """
        if page_identifier.isdigit():
            page_id = page_identifier
        else:
            page = await self.get_page_by_title(page_identifier)
            page_id = page.get("page", {}).get("id")

        if not page_id:
            raise ToolExecutionError(message=f"No page found with identifier: '{page_identifier}'")

        return page_id

    async def get_space_id(self, space_identifier: str) -> str:
        """Get the ID of a space from its identifier.

        Args:
            space_identifier: The identifier of the space to get. Can be a space's ID or title.
        """
        if space_identifier.isdigit():
            space_id = space_identifier
        else:
            space = await self.get_space_by_key(space_identifier)
            space_id = space.get("space", {}).get("id")
        return space_id

    def create_space_tree(self, space: dict) -> dict:
        """Create the initial tree structure for a space.

        Args:
            space: The transformed space object

        Returns:
            A dictionary representing the root of the space hierarchy tree without any children
        """
        space_internal = space.get("space", {})
        return {
            "key": space_internal.get("key"),
            "id": space_internal.get("id"),
            "type": "space",
            "url": space_internal.get("url"),
            "children": [],
        }

    def convert_root_pages_to_tree_nodes(self, pages: list) -> list:
        """Convert root pages of a space to tree nodes.

        Args:
            pages: List of page objects from the API

        Returns:
            A list of tree nodes representing the root pages
        """
        return [
            {
                "title": page.get("title"),
                "id": page.get("id"),
                "type": "page",
                "url": page.get("url"),
                "children": [],
            }
            for page in pages
        ]

    async def process_page_descendants(self, root_children: list, base_url: str) -> None:
        """Process descendants for each page and build the hierarchy.

        Args:
            root_children: The root children of the space
            base_url: The base URL for the Confluence space

        Returns:
            None (modifies root_children in place)
        """
        descendent_params = {"limit": 250, "depth": 5}

        for i, child in enumerate(root_children):
            page_id = child["id"]
            descendants = await self.get(f"pages/{page_id}/descendants", params=descendent_params)

            # Transform descendants into our desired format
            transformed_children = []
            for descendant in descendants.get("results", []):
                transformed_child = {
                    "title": descendant.get("title"),
                    "id": descendant.get("id"),
                    "type": descendant.get("type"),
                    "parent_id": page_id
                    if descendant.get("parentId") is None
                    else descendant.get("parentId"),
                    "parent_type": descendant.get("parentType", "TODO"),
                    "url": build_child_url(base_url, descendant),
                    "children": [],
                    "status": descendant.get("status"),
                }
                transformed_children.append(transformed_child)

            # Build the hierarchy for the current root page
            build_hierarchy(transformed_children, page_id, root_children[i])



import logging
import re
from base64 import urlsafe_b64decode, urlsafe_b64encode
from datetime import datetime, timedelta
from email.message import EmailMessage
from email.mime.text import MIMEText
from enum import Enum
from typing import Any

from arcade_tdk import ToolContext
from bs4 import BeautifulSoup
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build

from arcade_gmail.enums import (
    GmailAction,
    GmailReplyToWhom,
)
from arcade_gmail.exceptions import GmailServiceError, GmailToolError

logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)

logger = logging.getLogger(__name__)


class DateRange(Enum):
    TODAY = "today"
    YESTERDAY = "yesterday"
    LAST_7_DAYS = "last_7_days"
    LAST_30_DAYS = "last_30_days"
    THIS_MONTH = "this_month"
    LAST_MONTH = "last_month"
    THIS_YEAR = "this_year"

    def to_date_query(self) -> str:
        today = datetime.now()
        result = "after:"
        comparison_date = today

        if self == DateRange.YESTERDAY:
            comparison_date = today - timedelta(days=1)
        elif self == DateRange.LAST_7_DAYS:
            comparison_date = today - timedelta(days=7)
        elif self == DateRange.LAST_30_DAYS:
            comparison_date = today - timedelta(days=30)
        elif self == DateRange.THIS_MONTH:
            comparison_date = today.replace(day=1)
        elif self == DateRange.LAST_MONTH:
            comparison_date = (today.replace(day=1) - timedelta(days=1)).replace(day=1)
        elif self == DateRange.THIS_YEAR:
            comparison_date = today.replace(month=1, day=1)
        elif self == DateRange.LAST_MONTH:
            comparison_date = (today.replace(month=1, day=1) - timedelta(days=1)).replace(
                month=1, day=1
            )

        return result + comparison_date.strftime("%Y/%m/%d")


def build_email_message(
    recipient: str,
    subject: str,
    body: str,
    cc: list[str] | None = None,
    bcc: list[str] | None = None,
    replying_to: dict[str, Any] | None = None,
    action: GmailAction = GmailAction.SEND,
) -> dict[str, Any]:
    if replying_to:
        body = build_reply_body(body, replying_to)

    message: EmailMessage | MIMEText

    if action == GmailAction.SEND:
        message = EmailMessage()
        message.set_content(body)
    elif action == GmailAction.DRAFT:
        message = MIMEText(body)

    message["To"] = recipient
    message["Subject"] = subject

    if cc:
        message["Cc"] = ",".join(cc)
    if bcc:
        message["Bcc"] = ",".join(bcc)
    if replying_to:
        message["In-Reply-To"] = replying_to["header_message_id"]
        message["References"] = f"{replying_to['header_message_id']}, {replying_to['references']}"

    encoded_message = urlsafe_b64encode(message.as_bytes()).decode()

    data = {"raw": encoded_message}

    if replying_to:
        data["threadId"] = replying_to["thread_id"]

    return data


def _build_gmail_service(context: ToolContext) -> Any:
    """
    Private helper function to build and return the Gmail service client.

    Args:
        context (ToolContext): The context containing authorization details.

    Returns:
        googleapiclient.discovery.Resource: An authorized Gmail API service instance.
    """
    try:
        credentials = Credentials(
            context.authorization.token
            if context.authorization and context.authorization.token
            else ""
        )
    except Exception as e:
        raise GmailServiceError(message="Failed to build Gmail service.", developer_message=str(e))

    return build("gmail", "v1", credentials=credentials)


def build_gmail_query_string(
    sender: str | None = None,
    recipient: str | None = None,
    subject: str | None = None,
    body: str | None = None,
    date_range: DateRange | None = None,
    label: str | None = None,
) -> str:
    """Helper function to build a query string
    for Gmail list_emails_by_header and search_threads tools.
    """
    query = []
    if sender:
        query.append(f"from:{sender}")
    if recipient:
        query.append(f"to:{recipient}")
    if subject:
        query.append(f"subject:{subject}")
    if body:
        query.append(body)
    if date_range:
        query.append(date_range.to_date_query())
    if label:
        query.append(f"label:{label}")
    return " ".join(query)


def get_label_ids(service: Any, label_names: list[str]) -> dict[str, str]:
    """
    Retrieve label IDs for given label names.
    Returns a dictionary mapping label names to their IDs.

    Args:
        service: Authenticated Gmail API service instance.
        label_names: List of label names to retrieve IDs for.

    Returns:
        A dictionary mapping found label names to their corresponding IDs.
    """
    try:
        # Fetch all existing labels from Gmail
        labels = service.users().labels().list(userId="me").execute().get("labels", [])
    except Exception as e:
        raise GmailToolError(message="Failed to list labels.", developer_message=str(e)) from e

    # Create a mapping from label names to their IDs
    label_id_map = {label["name"]: label["id"] for label in labels}

    found_labels = {}
    for name in label_names:
        label_id = label_id_map.get(name)
        if label_id:
            found_labels[name] = label_id
        else:
            logger.warning(f"Label '{name}' does not exist")

    return found_labels


def fetch_messages(service: Any, query_string: str, limit: int) -> list[dict[str, Any]]:
    """
    Helper function to fetch messages from Gmail API for the list_emails_by_header tool.
    """
    response = (
        service.users()
        .messages()
        .list(userId="me", q=query_string, maxResults=limit or 100)
        .execute()
    )
    return response.get("messages", [])  # type: ignore[no-any-return]


def remove_none_values(params: dict) -> dict:
    """
    Remove None values from a dictionary.
    :param params: The dictionary to clean
    :return: A new dictionary with None values removed
    """
    return {k: v for k, v in params.items() if v is not None}


def build_reply_recipients(
    replying_to: dict[str, Any], current_user_email_address: str, reply_to_whom: GmailReplyToWhom
) -> str:
    if reply_to_whom == GmailReplyToWhom.ONLY_THE_SENDER:
        recipients = [replying_to["from"]]
    elif reply_to_whom == GmailReplyToWhom.EVERY_RECIPIENT:
        recipients = [replying_to["from"], *replying_to["to"].split(",")]
    else:
        raise ValueError(f"Unsupported reply_to_whom value: {reply_to_whom}")

    recipients = [
        email_address.strip()
        for email_address in recipients
        if email_address.strip().lower() != current_user_email_address.lower().strip()
    ]

    return ", ".join(recipients)


def get_draft_url(draft_id: str) -> str:
    return f"https://mail.google.com/mail/u/0/#drafts/{draft_id}"


def get_sent_email_url(sent_email_id: str) -> str:
    return f"https://mail.google.com/mail/u/0/#sent/{sent_email_id}"


def get_email_details(service: Any, messages: list[dict[str, Any]]) -> list[dict[str, Any]]:
    """
    Retrieves full message data for each message ID in the given list and extracts email details.

    :param service: Authenticated Gmail API service instance.
    :param messages: A list of dictionaries, each representing a message with an 'id' key.
    :return: A list of dictionaries, each containing parsed email details.
    """

    emails = []
    for msg in messages:
        try:
            # Fetch the full message data from Gmail using the message ID
            email_data = service.users().messages().get(userId="me", id=msg["id"]).execute()
            # Parse the raw email data into a structured form
            email_details = parse_plain_text_email(email_data)
            # Only add the details if parsing was successful
            if email_details:
                emails.append(email_details)
        except Exception as e:
            # Log any errors encountered while trying to fetch or parse a message
            raise GmailToolError(
                message=f"Error reading email {msg['id']}.", developer_message=str(e)
            )
    return emails


def get_email_in_trash_url(email_id: str) -> str:
    return f"https://mail.google.com/mail/u/0/#trash/{email_id}"


def parse_draft_email(draft_email_data: dict[str, Any]) -> dict[str, str]:
    """
    Parse draft email data and extract relevant information.

    Args:
        draft_email_data (Dict[str, Any]): Raw draft email data from Gmail API.

    Returns:
        dict[str, str]: Parsed draft email details
    """
    message = draft_email_data.get("message", {})
    payload = message.get("payload", {})
    headers = {d["name"].lower(): d["value"] for d in payload.get("headers", [])}

    body_data = _get_email_plain_text_body(payload)

    return {
        "id": draft_email_data.get("id", ""),
        "thread_id": draft_email_data.get("threadId", ""),
        "from": headers.get("from", ""),
        "date": headers.get("internaldate", ""),
        "subject": headers.get("subject", ""),
        "body": _clean_email_body(body_data) if body_data else "",
    }


def _clean_email_body(body: str | None) -> str:
    """
    Remove HTML tags and clean up email body text while preserving most content.

    Args:
        body (str): The raw email body text.

    Returns:
        str: Cleaned email body text.
    """
    if not body:
        return ""

    try:
        # Remove HTML tags using BeautifulSoup
        soup = BeautifulSoup(body, "html.parser")
        text = soup.get_text(separator=" ")

        # Clean up the text
        cleaned_text = _clean_text(text)

        return cleaned_text.strip()
    except Exception:
        logger.exception("Error cleaning email body")
        return body


def _get_email_plain_text_body(payload: dict[str, Any]) -> str | None:
    """
    Extract email body from payload, handling 'multipart/alternative' parts.

    Args:
        payload (Dict[str, Any]): Email payload data.

    Returns:
        str | None: Decoded email body or None if not found.
    """
    # Direct body extraction
    if "body" in payload and payload["body"].get("data"):
        return _clean_email_body(urlsafe_b64decode(payload["body"]["data"]).decode())

    # Handle multipart and alternative parts
    return _clean_email_body(_extract_plain_body(payload.get("parts", [])))


def _extract_plain_body(parts: list) -> str | None:
    """
    Recursively extract the email body from parts, handling both plain text and HTML.

    Args:
        parts (List[Dict[str, Any]]): List of email parts.

    Returns:
        str | None: Decoded and cleaned email body or None if not found.
    """
    for part in parts:
        mime_type = part.get("mimeType")

        if mime_type == "text/plain" and "data" in part.get("body", {}):
            return urlsafe_b64decode(part["body"]["data"]).decode()

        elif mime_type.startswith("multipart/"):
            subparts = part.get("parts", [])
            body = _extract_plain_body(subparts)
            if body:
                return body

    return _extract_html_body(parts)


def _extract_html_body(parts: list) -> str | None:
    """
    Recursively extract the email body from parts, handling only HTML.

    Args:
        parts (List[Dict[str, Any]]): List of email parts.

    Returns:
        str | None: Decoded and cleaned email body or None if not found.
    """
    for part in parts:
        mime_type = part.get("mimeType")

        if mime_type == "text/html" and "data" in part.get("body", {}):
            html_content = urlsafe_b64decode(part["body"]["data"]).decode()
            return html_content

        elif mime_type.startswith("multipart/"):
            subparts = part.get("parts", [])
            body = _extract_html_body(subparts)
            if body:
                return body

    return None


def _clean_text(text: str) -> str:
    """
    Clean up the text while preserving most content.

    Args:
        text (str): The input text.

    Returns:
        str: Cleaned text.
    """
    # Replace multiple newlines with a single newline
    text = re.sub(r"\n+", "\n", text)

    # Replace multiple spaces with a single space
    text = re.sub(r"\s+", " ", text)

    # Remove leading/trailing whitespace from each line
    text = "\n".join(line.strip() for line in text.split("\n"))

    return text


def parse_plain_text_email(email_data: dict[str, Any]) -> dict[str, Any]:
    """
    Parse email data and extract relevant information.
    Only returns the plain text body.

    Args:
        email_data (dict[str, Any]): Raw email data from Gmail API.

    Returns:
        dict[str, str]: Parsed email details
    """
    payload = email_data.get("payload", {})
    headers = {d["name"].lower(): d["value"] for d in payload.get("headers", [])}

    body_data = _get_email_plain_text_body(payload)

    email_details = {
        "id": email_data.get("id", ""),
        "thread_id": email_data.get("threadId", ""),
        "label_ids": email_data.get("labelIds", []),
        "history_id": email_data.get("historyId", ""),
        "snippet": email_data.get("snippet", ""),
        "to": headers.get("to", ""),
        "cc": headers.get("cc", ""),
        "from": headers.get("from", ""),
        "reply_to": headers.get("reply-to", ""),
        "in_reply_to": headers.get("in-reply-to", ""),
        "references": headers.get("references", ""),
        "header_message_id": headers.get("message-id", ""),
        "date": headers.get("date", ""),
        "subject": headers.get("subject", ""),
        "body": body_data or "",
    }

    return email_details


def build_reply_body(body: str, replying_to: dict[str, Any]) -> str:
    attribution = f"On {replying_to['date']}, {replying_to['from']} wrote:"
    lines = replying_to["plain_text_body"].split("\n")
    quoted_plain = "\n".join([f"> {line}" for line in lines])
    return f"{body}\n\n{attribution}\n\n{quoted_plain}"


def parse_multipart_email(email_data: dict[str, Any]) -> dict[str, Any]:
    """
    Parse email data and extract relevant information.
    Returns the plain text and HTML body along with the images.

    Args:
        email_data (Dict[str, Any]): Raw email data from Gmail API.

    Returns:
        dict[str, Any]: Parsed email details
    """

    payload = email_data.get("payload", {})
    headers = {d["name"].lower(): d["value"] for d in payload.get("headers", [])}

    # Extract different parts of the email
    plain_text_body = _get_email_plain_text_body(payload)
    html_body = _get_email_html_body(payload)

    email_details = {
        "id": email_data.get("id", ""),
        "thread_id": email_data.get("threadId", ""),
        "label_ids": email_data.get("labelIds", []),
        "history_id": email_data.get("historyId", ""),
        "snippet": email_data.get("snippet", ""),
        "to": headers.get("to", ""),
        "cc": headers.get("cc", ""),
        "from": headers.get("from", ""),
        "reply_to": headers.get("reply-to", ""),
        "in_reply_to": headers.get("in-reply-to", ""),
        "references": headers.get("references", ""),
        "header_message_id": headers.get("message-id", ""),
        "date": headers.get("date", ""),
        "subject": headers.get("subject", ""),
        "plain_text_body": plain_text_body or _clean_email_body(html_body),
        "html_body": html_body or "",
    }

    return email_details


def _get_email_html_body(payload: dict[str, Any]) -> str | None:
    """
    Extract email html body from payload, handling 'multipart/alternative' parts.

    Args:
        payload (Dict[str, Any]): Email payload data.

    Returns:
        str | None: Decoded email body or None if not found.
    """
    # Direct body extraction
    if "body" in payload and payload["body"].get("data"):
        return urlsafe_b64decode(payload["body"]["data"]).decode()

    # Handle multipart and alternative parts
    return _extract_html_body(payload.get("parts", []))



import asyncio
import json
from collections.abc import Awaitable
from datetime import datetime
from typing import Any, Callable, TypeVar, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import RetryableToolError, ToolExecutionError

from arcade_asana.constants import (
    ASANA_MAX_TIMEOUT_SECONDS,
    MAX_PROJECTS_TO_SCAN_BY_NAME,
    MAX_TAGS_TO_SCAN_BY_NAME,
    TASK_OPT_FIELDS,
    SortOrder,
    TaskSortBy,
)
from arcade_asana.exceptions import PaginationTimeoutError

ToolResponse = TypeVar("ToolResponse", bound=dict[str, Any])


def remove_none_values(data: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in data.items() if v is not None}


def validate_date_format(name: str, date_str: str | None) -> None:
    if not date_str:
        return

    try:
        datetime.strptime(date_str, "%Y-%m-%d")
    except ValueError:
        raise ToolExecutionError(f"Invalid {name} date format. Use the format YYYY-MM-DD.")


def build_task_search_query_params(
    keywords: str | None,
    completed: bool | None,
    assignee_id: str | None,
    project_id: str | None,
    team_id: str | None,
    tag_ids: list[str] | None,
    due_on: str | None,
    due_on_or_after: str | None,
    due_on_or_before: str | None,
    start_on: str | None,
    start_on_or_after: str | None,
    start_on_or_before: str | None,
    limit: int,
    sort_by: TaskSortBy,
    sort_order: SortOrder,
) -> dict[str, Any]:
    query_params: dict[str, Any] = {
        "text": keywords,
        "opt_fields": TASK_OPT_FIELDS,
        "sort_by": sort_by.value,
        "sort_ascending": sort_order == SortOrder.ASCENDING,
        "limit": limit,
    }
    if completed is not None:
        query_params["completed"] = completed
    if assignee_id:
        query_params["assignee.any"] = assignee_id
    if project_id:
        query_params["projects.any"] = project_id
    if team_id:
        query_params["team.any"] = team_id
    if tag_ids:
        query_params["tags.any"] = ",".join(tag_ids)

    query_params = add_task_search_date_params(
        query_params,
        due_on,
        due_on_or_after,
        due_on_or_before,
        start_on,
        start_on_or_after,
        start_on_or_before,
    )

    return query_params


def add_task_search_date_params(
    query_params: dict[str, Any],
    due_on: str | None,
    due_on_or_after: str | None,
    due_on_or_before: str | None,
    start_on: str | None,
    start_on_or_after: str | None,
    start_on_or_before: str | None,
) -> dict[str, Any]:
    """
    Builds the date-related query parameters for task search.

    If a date is provided, it will be added to the query parameters. If not, it will be ignored.
    """
    if due_on:
        query_params["due_on"] = due_on
    if due_on_or_after:
        query_params["due_on.after"] = due_on_or_after
    if due_on_or_before:
        query_params["due_on.before"] = due_on_or_before
    if start_on:
        query_params["start_on"] = start_on
    if start_on_or_after:
        query_params["start_on.after"] = start_on_or_after
    if start_on_or_before:
        query_params["start_on.before"] = start_on_or_before

    return query_params


async def handle_new_task_associations(
    context: ToolContext,
    parent_task_id: str | None,
    project: str | None,
    workspace_id: str | None,
) -> tuple[str | None, str | None, str | None]:
    """
    Handles the association of a new task to a parent task, project, or workspace.

    If no association is provided, it will try to find a workspace in the user's account.
    In case the user has only one workspace, it will use that workspace.
    Otherwise, it will raise an error.

    If a workspace_id is not provided, but a parent_task_id or a project_id is provided, it will try
    to find the workspace associated with the parent task or project.

    In each of the two cases explained above, if a workspace is found, the function will return this
    value, even if the workspace_id argument was None.

    Returns a tuple of (parent_task_id, project_id, workspace_id).
    """
    project_id, project_name = (None, None)

    if project:
        if project.isnumeric():
            project_id = project
        else:
            project_name = project

    if project_name:
        project_data = await get_project_by_name_or_raise_error(context, project_name)
        project_id = project_data["id"]
        workspace_id = project_data["workspace"]["id"]

    if not any([parent_task_id, project_id, workspace_id]):
        workspace_id = await get_unique_workspace_id_or_raise_error(context)

    if not workspace_id and parent_task_id:
        from arcade_asana.tools.tasks import get_task_by_id  # avoid circular imports

        response = await get_task_by_id(context, parent_task_id)
        workspace_id = response["task"]["workspace"]["id"]

    return parent_task_id, project_id, workspace_id


async def get_project_by_name_or_raise_error(
    context: ToolContext,
    project_name: str,
    max_items_to_scan: int = MAX_PROJECTS_TO_SCAN_BY_NAME,
) -> dict[str, Any]:
    response = await find_projects_by_name(
        context=context,
        names=[project_name],
        response_limit=100,
        max_items_to_scan=max_items_to_scan,
        return_projects_not_matched=True,
    )

    if not response["matches"]["projects"]:
        projects = response["not_matched"]["projects"]
        projects = [{"name": project["name"], "id": project["id"]} for project in projects]
        message = (
            f"Project with name '{project_name}' was not found. The search scans up to "
            f"{max_items_to_scan} projects. If the user account has a larger number of projects, "
            "it's possible that it exists, but the search didn't find it."
        )
        additional_prompt = f"Projects available: {json.dumps(projects)}"
        raise RetryableToolError(
            message=message,
            developer_message=f"{message} {additional_prompt}",
            additional_prompt_content=additional_prompt,
        )

    elif response["matches"]["count"] > 1:
        projects = [
            {"name": project["name"], "id": project["id"]}
            for project in response["matches"]["projects"]
        ]
        message = "Multiple projects found with the same name. Please provide a project ID instead."
        additional_prompt = f"Projects matching the name '{project_name}': {json.dumps(projects)}"
        raise RetryableToolError(
            message=message,
            developer_message=message,
            additional_prompt_content=additional_prompt,
        )

    return cast(dict, response["matches"]["projects"][0])


async def handle_new_task_tags(
    context: ToolContext,
    tags: list[str] | None,
    workspace_id: str | None,
) -> list[str] | None:
    if not tags:
        return None

    tag_ids = []
    tag_names = []
    for tag in tags:
        if tag.isnumeric():
            tag_ids.append(tag)
        else:
            tag_names.append(tag)

    if tag_names:
        response = await find_tags_by_name(context, tag_names)
        tag_ids.extend([tag["id"] for tag in response["matches"]["tags"]])

        if response["not_found"]["tags"]:
            from arcade_asana.tools.tags import create_tag  # avoid circular imports

            responses = await asyncio.gather(*[
                create_tag(context, name=name, workspace_id=workspace_id)
                for name in response["not_found"]["tags"]
            ])
            tag_ids.extend([response["tag"]["id"] for response in responses])

    return tag_ids


async def get_tag_ids(
    context: ToolContext,
    tags: list[str] | None,
    max_items_to_scan: int = MAX_TAGS_TO_SCAN_BY_NAME,
) -> list[str] | None:
    """
    Returns the IDs of the tags provided in the tags list, which can be either tag IDs or tag names.

    If the tags list is empty, it returns None.
    """
    tag_ids = []
    tag_names = []

    if tags:
        for tag in tags:
            if tag.isnumeric():
                tag_ids.append(tag)
            else:
                tag_names.append(tag)

    if tag_names:
        searched_tags = await find_tags_by_name(
            context=context,
            names=tag_names,
            max_items_to_scan=max_items_to_scan,
        )

        if searched_tags["not_found"]["tags"]:
            tag_names_not_found = ", ".join(searched_tags["not_found"]["tags"])
            raise ToolExecutionError(
                f"Tags not found: {tag_names_not_found}. The search scans up to "
                f"{max_items_to_scan} tags. If the user account has a larger number of tags, "
                "it's possible that the tags exist, but the search didn't find them."
            )

        tag_ids.extend([tag["id"] for tag in searched_tags["matches"]["tags"]])

    return tag_ids if tag_ids else None


async def paginate_tool_call(
    tool: Callable[[ToolContext, Any], Awaitable[ToolResponse]],
    context: ToolContext,
    response_key: str,
    max_items: int = 300,
    timeout_seconds: int = ASANA_MAX_TIMEOUT_SECONDS,
    next_page_token: str | None = None,
    **tool_kwargs: Any,
) -> list[ToolResponse]:
    results: list[ToolResponse] = []

    async def paginate_loop() -> None:
        nonlocal results
        keep_paginating = True

        if "limit" not in tool_kwargs:
            tool_kwargs["limit"] = 100

        while keep_paginating:
            response = await tool(context, **tool_kwargs)  # type: ignore[call-arg]
            results.extend(response[response_key])
            next_page = get_next_page(response)
            next_page_token = next_page["next_page_token"]
            if not next_page_token or len(results) >= max_items:
                keep_paginating = False
            else:
                tool_kwargs["next_page_token"] = next_page_token

    try:
        await asyncio.wait_for(paginate_loop(), timeout=timeout_seconds)
    except TimeoutError:
        raise PaginationTimeoutError(timeout_seconds, tool.__tool_name__)  # type: ignore[attr-defined]
    else:
        return results


async def get_unique_workspace_id_or_raise_error(context: ToolContext) -> str:
    # Importing here to avoid circular imports
    from arcade_asana.tools.workspaces import list_workspaces

    workspaces = await list_workspaces(context)
    if len(workspaces["workspaces"]) == 1:
        return cast(str, workspaces["workspaces"][0]["id"])
    else:
        message = "User has multiple workspaces. Please provide a workspace ID."
        additional_prompt = f"Workspaces available: {json.dumps(workspaces['workspaces'])}"
        raise RetryableToolError(
            message=message,
            developer_message=message,
            additional_prompt_content=additional_prompt,
        )


async def find_projects_by_name(
    context: ToolContext,
    names: list[str],
    team_id: list[str] | None = None,
    response_limit: int = 100,
    max_items_to_scan: int = MAX_PROJECTS_TO_SCAN_BY_NAME,
    return_projects_not_matched: bool = False,
) -> dict[str, Any]:
    """Find projects by name using exact match

    This function will paginate the list_projects tool call and return the projects that match
    the names provided. If the names provided are not found, it will return the names searched for
    that did not match any projects.

    If return_projects_not_matched is True, it will also return the projects that were scanned,
    but did not match any of the names searched for.

    Args:
        context: The tool context to use in the list_projects tool call.
        names: The names of the projects to search for.
        team_id: The ID of the team to search for projects in.
        response_limit: The maximum number of matched projects to return.
        max_items_to_scan: The maximum number of projects to scan while looking for matches.
        return_projects_not_matched: Whether to return the projects that were scanned, but did not
            match any of the names searched for.
    """
    from arcade_asana.tools.projects import list_projects  # avoid circular imports

    names_lower = {name.casefold() for name in names}

    projects = await paginate_tool_call(
        tool=list_projects,
        context=context,
        response_key="projects",
        max_items=max_items_to_scan,
        timeout_seconds=15,
        team_id=team_id,
    )

    matches: list[dict[str, Any]] = []
    not_matched: list[str] = []

    for project in projects:
        project_name_lower = project["name"].casefold()
        if len(matches) >= response_limit:
            break
        if project_name_lower in names_lower:
            matches.append(project)
            names_lower.remove(project_name_lower)
        else:
            not_matched.append(project)

    not_found = [name for name in names if name.casefold() in names_lower]

    response = {
        "matches": {
            "projects": matches,
            "count": len(matches),
        },
        "not_found": {
            "names": not_found,
            "count": len(not_found),
        },
    }

    if return_projects_not_matched:
        response["not_matched"] = {
            "projects": not_matched,
            "count": len(not_matched),
        }

    return response


async def find_tags_by_name(
    context: ToolContext,
    names: list[str],
    workspace_id: list[str] | None = None,
    response_limit: int = 100,
    max_items_to_scan: int = MAX_TAGS_TO_SCAN_BY_NAME,
    return_tags_not_matched: bool = False,
) -> dict[str, Any]:
    """Find tags by name using exact match

    This function will paginate the list_tags tool call and return the tags that match the names
    provided. If the names provided are not found, it will return the names searched for that did
    not match any tags.

    If return_tags_not_matched is True, it will also return the tags that were scanned, but did not
    match any of the names searched for.

    Args:
        context: The tool context to use in the list_tags tool call.
        names: The names of the tags to search for.
        workspace_id: The ID of the workspace to search for tags in.
        response_limit: The maximum number of matched tags to return.
        max_items_to_scan: The maximum number of tags to scan while looking for matches.
        return_tags_not_matched: Whether to return the tags that were scanned, but did not match
            any of the names searched for.
    """
    from arcade_asana.tools.tags import list_tags  # avoid circular imports

    names_lower = {name.casefold() for name in names}

    tags = await paginate_tool_call(
        tool=list_tags,
        context=context,
        response_key="tags",
        max_items=max_items_to_scan,
        timeout_seconds=15,
        workspace_id=workspace_id,
    )

    matches: list[dict[str, Any]] = []
    not_matched: list[str] = []
    for tag in tags:
        tag_name_lower = tag["name"].casefold()
        if len(matches) >= response_limit:
            break
        if tag_name_lower in names_lower:
            matches.append(tag)
            names_lower.remove(tag_name_lower)
        else:
            not_matched.append(tag["name"])

    not_found = [name for name in names if name.casefold() in names_lower]

    response = {
        "matches": {
            "tags": matches,
            "count": len(matches),
        },
        "not_found": {
            "tags": not_found,
            "count": len(not_found),
        },
    }

    if return_tags_not_matched:
        response["not_matched"] = {
            "tags": not_matched,
            "count": len(not_matched),
        }

    return response


def get_next_page(response: dict[str, Any]) -> dict[str, Any]:
    try:
        token = response["next_page"]["offset"]
    except (KeyError, TypeError):
        token = None

    return {"has_more_pages": token is not None, "next_page_token": token}



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_math
from arcade_math.tools.arithmetic import (
    add,
    divide,
    mod,
    multiply,
    subtract,
    sum_list,
    sum_range,
)
from arcade_math.tools.exponents import (
    log,
    power,
)
from arcade_math.tools.miscellaneous import (
    abs_val,
    factorial,
    sqrt,
)
from arcade_math.tools.rational import (
    gcd,
    lcm,
)
from arcade_math.tools.rounding import (
    ceil,
    floor,
    round_num,
)
from arcade_math.tools.statistics import (
    avg,
    median,
)
from arcade_math.tools.trigonometry import (
    deg_to_rad,
    rad_to_deg,
)

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_math)


@tool_eval()
def math_eval_suite():
    suite = EvalSuite(
        name="Math Tools Evaluation",
        system_message="You're an AI assistant with access to math tools. Use them to help the user with their math-related tasks.",
        catalog=catalog,
        rubric=rubric,
    )

    list_param = ["1", "2", "3", "4", "5"]
    funcs_to_expression_and_params = [
        # unary
        (sqrt, "What's the square root of {a}?", {"a": "25"}),
        (abs_val, "What's the absolute value of {a}?", {"a": "-10"}),
        (factorial, "What's the factorial of {a}?", {"a": "5"}),
        (deg_to_rad, "Convert {degrees} from degrees to radians", {"degrees": "180"}),
        (rad_to_deg, "Convert {radians} from radias to degrees", {"radians": "3.14"}),
        (ceil, "Compute the ceiling of {a}", {"a": "3.14"}),
        (floor, "Compute the floor of {a}", {"a": "3.14"}),
        # binary
        (add, "Add {a} and {b}", {"a": "12345", "b": "987654321"}),
        (subtract, "Subtract {b} from {a}", {"a": "987654321", "b": "12345"}),
        (multiply, "Multiply {a} and {b}", {"a": "12345", "b": "567890"}),
        (divide, "What is {a} divided by {b}?", {"a": "1234123479", "b": "123"}),
        (
            sum_range,
            "What's the sum of all numbers from {start} to {end}?",
            {"start": "10", "end": "345"},
        ),
        (mod, "What's the remainder of dividing {a} by {b}?", {"a": "234", "b": "17"}),
        (power, "Raise {a} to the power of {b}", {"a": "2", "b": "8"}),
        (log, "What's the logarithm of {a} with base {base}?", {"a": "8", "base": "2"}),
        (
            round_num,
            "Round {value} to {ndigits} decimal places",
            {"value": "12.23746234", "ndigits": "3"},
        ),
        (gcd, "Find the greatest common divisor of {a} and {b}", {"a": "50", "b": "10"}),
        (lcm, "FInd the least common multiple of {a} and {b}", {"a": "7", "b": "13"}),
        # n-nary
        (
            sum_list,
            f"Calculate the sum of these numbers: {' '.join(list_param)}",
            {"numbers": list_param},
        ),
        (
            avg,
            f"Find the average of these numbers: {' '.join(list_param)}",
            {"numbers": list_param},
        ),
        (
            median,
            f"Find the median of these numbers: {' '.join(list_param)}",
            {"numbers": list_param},
        ),
    ]

    for func, expression, params in funcs_to_expression_and_params:
        parametrized_expression = expression.format(**params)
        num_params = len(params)
        suite.add_case(
            name=parametrized_expression,
            user_message=parametrized_expression,
            expected_tool_calls=[
                ExpectedToolCall(
                    func=func,
                    args=params,
                )
            ],
            rubric=rubric,
            critics=[BinaryCritic(critic_field=param, weight=1.0 / num_params) for param in params],
        )

    return suite



from collections.abc import Collection
from typing import Any

from arcade_evals import DatetimeCritic


class DatetimeOrNoneCritic(DatetimeCritic):
    """
    A critic that evaluates the closeness of datetime values within a specified tolerance or whether
    it's a None value.

    Attributes:
        tolerance: Acceptable timedelta between expected and actual datetimes.
        max_difference: Maximum timedelta for a partial score.
    """

    def evaluate(self, expected: Any, actual: Any) -> dict[str, Any]:
        if actual is None:
            return {"match": True, "score": self.weight}
        return super().evaluate(expected, actual)


class AnyDatetimeCritic(DatetimeCritic):
    """
    A critic that evaluates the closeness of datetime values within a list of expected values.
    """

    def evaluate(self, expected: Any, actual: Any) -> dict[str, Any]:
        if not isinstance(expected, Collection):
            expected = [expected]
        for expected_value in expected:
            critic = DatetimeCritic(
                critic_field=self.critic_field,
                weight=self.weight,
                tolerance=self.tolerance,
                max_difference=self.max_difference,
            )
            result = critic.evaluate(expected_value, actual)
            if result["match"]:
                return result
        return {"match": False, "score": 0}



import pytest
from arcade_tdk.errors import ToolExecutionError

from {{ package_name }}.tools.hello import say_hello


def test_hello() -> None:
    assert say_hello("developer") == "Hello, developer!"


def test_hello_raises_error() -> None:
    with pytest.raises(ToolExecutionError):
        say_hello(1)




import decimal
import math
from decimal import Decimal
from typing import Annotated

from arcade_tdk import tool

decimal.getcontext().prec = 100


@tool
def abs_val(
    a: Annotated[str, "The number as a string"],
) -> Annotated[str, "The absolute value of the number as a string"]:
    """
    Calculate the absolute value of a number
    """
    # Use Decimal for arbitrary precision
    return str(abs(Decimal(a)))


@tool
def factorial(
    a: Annotated[str, "The non-negative integer to compute the factorial for as a string"],
) -> Annotated[str, "The factorial of the number as a string"]:
    """
    Compute the factorial of a non-negative integer
    Returns "1" for "0"
    """
    return str(math.factorial(int(a)))


@tool
def sqrt(
    a: Annotated[str, "The number to square root as a string"],
) -> Annotated[str, "The square root of the number as a string"]:
    """
    Get the square root of a number
    """
    # Use Decimal for arbitrary precision
    a_decimal = Decimal(a)
    return str(a_decimal.sqrt())



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

from arcade_spotify.tools.tracks import get_track_from_id

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)

catalog = ToolCatalog()
catalog.add_tool(get_track_from_id, "Spotify")

# This history is a conversation where the user asks for the currently playing song, and then asks for information about it.
history = [
    {"role": "user", "content": "what song is playing rn"},
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_fqZHT1DNcil1pbIv6UbOWVq9",
                "type": "function",
                "function": {"name": "Spotify_GetCurrentlyPlaying", "arguments": "{}"},
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"album_artists":["Gajaka"],"album_id":"7b5h7MJ70KC92FrwSvRaQi","album_name":"Aerial","album_spotify_url":"https://open.spotify.com/album/7b5h7MJ70KC92FrwSvRaQi","currently_playing_type":"track","is_playing":true,"progress_ms":96962,"track_artists":["Gajaka"],"track_id":"03gaqN3aWm9TQxuHay0G8R","track_name":"Aerial"}',
        "tool_call_id": "call_fqZHT1DNcil1pbIv6UbOWVq9",
        "name": "Spotify_GetCurrentlyPlaying",
    },
    {
        "role": "assistant",
        "content": 'The song currently playing is "Aerial" by Gajaka. You can listen to it on Spotify [here](https://open.spotify.com/album/7b5h7MJ70KC92FrwSvRaQi).',
    },
]


@tool_eval()
def spotify_tracks_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Spotify "track" tools."""
    suite = EvalSuite(
        name="Spotify Tools Evaluation",
        system_message="You are an AI assistant that can manage Spotify using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get information about a track",
        user_message="can you get more info about that song",
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_track_from_id,
                args={
                    "track_id": "03gaqN3aWm9TQxuHay0G8R",
                },
            )
        ],
        critics=[
            BinaryCritic(critic_field="track_id", weight=1.0),
        ],
        additional_messages=history,
    )

    return suite



import functools
from collections.abc import Callable
from typing import Any

from arcade_tdk import ToolContext
from googleapiclient.errors import HttpError

from arcade_google_sheets.file_picker import generate_google_file_picker_url


def with_filepicker_fallback(func: Callable[..., Any]) -> Callable[..., Any]:
    """ """

    @functools.wraps(func)
    async def async_wrapper(context: ToolContext, *args: Any, **kwargs: Any) -> Any:
        try:
            return await func(context, *args, **kwargs)
        except HttpError as e:
            if e.status_code in [403, 404]:
                file_picker_response = generate_google_file_picker_url(context)
                return file_picker_response
            raise

    return async_wrapper



from arcade_evals import (
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_google
from arcade_google.tools import generate_google_file_picker_url

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.9,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_google)

history_1 = [
    {
        "role": "user",
        "content": "get the contents of my spreadsheet named 'How to teach an old dog a new trick'",
    },
    {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {
                "id": "call_errutwjcpZLiHu8LKip6XKLq",
                "type": "function",
                "function": {
                    "name": "Google_SearchDocuments",
                    "arguments": '{"document_contains":["How to teach an old dog a new trick"],"limit":5}',
                },
            }
        ],
    },
    {
        "role": "tool",
        "content": '{"documents":[],"documents_count":0}',
        "tool_call_id": "call_errutwjcpZLiHu8LKip6XKLq",
        "name": "Google_SearchDocuments",
    },
    {
        "role": "assistant",
        "content": "It seems that I don't have access to your spreadsheet named \"How to teach an old dog a new trick.\" If you need help retrieving the data, you can check your cloud storage or computer files where the spreadsheet might be saved, such as Google Sheets, Microsoft Excel, or other spreadsheet software. Let me know if there's anything specific you'd like assistance with regarding the content!",
    },
]

history_2 = [
    {"role": "system", "content": "Today is 2025-04-15, Tuesday."},
    {
        "role": "user",
        "content": "I got this error from you: \n❌ Arcade Chat failed with error: Error code: 500 - {'name':'non_retryable_tool_call_error', 'message': 'tool call failed - \nGoogle.InsertTextAtEndOfDocument failed: Error inexecution of GetDocumentById: Error in get_document_by_id: <HttpError 404 when\nrequestinghttps://docs.googleapis.com/v1/documents/192394932924?alt=json returned \"Requested entity was not found.\". Details: \n\"Requestedentity was not found.\">'}",
    },
    {
        "role": "assistant",
        "content": "The error message indicates that there was an attempt to retrieve a document with a specific ID \\(`192394932924`) from Google Docs, but the document could not be found. The HTTP 404 error code means that the document doesn't \\exist or is not accessible with the current permissions. Here's what you can do:\n\n1. **Check Document ID:** Make sure the \\document ID (`192394932924`) is correct. A typo or incorrect ID will lead to this error.\n\n2. **Permissions:** Ensure that you \\have permission to access the document. If the document is owned by someone else or is in a shared drive with restricted access,\\you might not be able to view it.\n\n3. **Document Existence:** Double-check that the document hasn't been deleted or moved to \the trash or a location you don't have access to.\n\nLet me know if there's anything else you'd like to address or any other \assistance you need!",
    },
]


@tool_eval()
def file_picker_eval_suite() -> EvalSuite:
    """Create an evaluation suite for Gmail tools."""
    suite = EvalSuite(
        name="File Picker Tools Evaluation",
        system_message="You are an AI assistant that can perform actions for the user using the provided tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get file picker url 1",
        user_message="I am certain that the spreadsheet exists",
        expected_tool_calls=[
            ExpectedToolCall(
                func=generate_google_file_picker_url,
                args={},
            )
        ],
        additional_messages=history_1,
    )

    suite.add_case(
        name="Get file picker url 2",
        user_message="Please help in anyway you can. I'm literally looking at the document right now in the browser.",
        expected_tool_calls=[
            ExpectedToolCall(func=generate_google_file_picker_url, args={}),
        ],
        additional_messages=history_2,
    )

    suite.add_case(
        name="Get file picker url 3",
        user_message="I want to give you permissions to more files in my Drive right now.",
        expected_tool_calls=[
            ExpectedToolCall(func=generate_google_file_picker_url, args={}),
        ],
    )

    return suite



import re
from typing import Any, cast

from arcade_tdk import ToolContext
from arcade_tdk.errors import ToolExecutionError
from serpapi import Client as SerpClient


def prepare_params(engine: str, **kwargs: Any) -> dict[str, Any]:
    """
    Prepares a parameters dictionary for the SerpAPI call.

    Parameters:
        engine: The engine name (e.g., "google", "google_finance").
        kwargs: Any additional parameters to include.

    Returns:
        A dictionary containing the base parameters plus any extras,
        excluding any parameters whose value is None.
    """
    params = {"engine": engine}
    params.update({k: v for k, v in kwargs.items() if v is not None})
    return params


def call_serpapi(context: ToolContext, params: dict) -> dict:
    """
    Execute a search query using the SerpAPI client and return the results as a dictionary.

    Args:
        context: The tool context containing required secrets.
        params: A dictionary of parameters for the SerpAPI search.

    Returns:
        The search results as a dictionary.
    """
    api_key = context.get_secret("SERP_API_KEY")
    client = SerpClient(api_key=api_key)
    try:
        search = client.search(params)
        return cast(dict[str, Any], search.as_dict())
    except Exception as e:
        # SerpAPI error messages sometimes contain the API key, so we need to sanitize it
        sanitized_e = re.sub(r"(api_key=)[^ &]+", r"\1***", str(e))
        raise ToolExecutionError(
            message="Failed to fetch search results",
            developer_message=sanitized_e,
        )



from typing import Annotated

from arcade_tdk import ToolContext, tool
from arcade_tdk.auth import Microsoft
from arcade_tdk.errors import ToolExecutionError

from arcade_microsoft.client import get_client
from arcade_microsoft.outlook_mail.message import Message, Recipient


@tool(requires_auth=Microsoft(scopes=["Mail.ReadWrite"]))
async def create_draft_email(
    context: ToolContext,
    subject: Annotated[str, "The subject of the draft email to create"],
    body: Annotated[str, "The body of the draft email to create"],
    to_recipients: Annotated[
        list[str], "The email addresses that will be the recipients of the draft email"
    ],
    cc_recipients: Annotated[
        list[str] | None,
        "The email addresses that will be the CC recipients of the draft email.",
    ] = None,
    bcc_recipients: Annotated[
        list[str] | None,
        "The email addresses that will be the BCC recipients of the draft email.",
    ] = None,
) -> Annotated[dict, "A dictionary containing the created email details"]:
    """Compose a new draft email in Outlook"""
    client = get_client(context.get_auth_token_or_empty())

    message = Message(
        subject=subject,
        body=body,
        to_recipients=[Recipient(email_address=email) for email in to_recipients],
        cc_recipients=[Recipient(email_address=email) for email in cc_recipients or []],
        bcc_recipients=[Recipient(email_address=email) for email in bcc_recipients or []],
        is_draft=True,
    ).to_sdk()

    response = await client.me.messages.post(message)
    draft_message = Message.from_sdk(response).to_dict()  # type: ignore[arg-type]

    return draft_message


@tool(requires_auth=Microsoft(scopes=["Mail.ReadWrite"]))
async def update_draft_email(
    context: ToolContext,
    message_id: Annotated[str, "The ID of the draft email to update"],
    subject: Annotated[
        str | None,
        "The new subject of the draft email. If provided, the existing subject will be overwritten",
    ] = None,
    body: Annotated[
        str | None,
        "The new body of the draft email. If provided, the existing body will be overwritten",
    ] = None,
    to_add: Annotated[list[str] | None, "Email addresses to add as 'To' recipients."] = None,
    to_remove: Annotated[
        list[str] | None,
        "Email addresses to remove from the current 'To' recipients.",
    ] = None,
    cc_add: Annotated[
        list[str] | None,
        "Email addresses to add as 'CC' recipients.",
    ] = None,
    cc_remove: Annotated[
        list[str] | None,
        "Email addresses to remove from the current 'CC' recipients.",
    ] = None,
    bcc_add: Annotated[
        list[str] | None,
        "Email addresses to add as 'BCC' recipients.",
    ] = None,
    bcc_remove: Annotated[
        list[str] | None,
        "Email addresses to remove from the current 'BCC' recipients.",
    ] = None,
) -> Annotated[dict, "A dictionary containing the updated email details"]:
    """Update an existing draft email in Outlook.

    This tool overwrites the subject and body of a draft email (if provided),
    and modifies its recipient lists by selectively adding or removing email addresses.

    This tool can update any un-sent email:
        - draft
        - reply-draft
        - reply-all draft
        - forward draft
    """
    client = get_client(context.get_auth_token_or_empty())

    # Get the draft email
    draft_email_sdk = await client.me.messages.by_message_id(message_id).get()

    if draft_email_sdk is None:
        raise ToolExecutionError(message=f"The draft email with ID {message_id} was not found.")

    # Update the draft email
    draft_email = Message.from_sdk(draft_email_sdk)
    draft_email.subject = subject if subject else draft_email.subject
    draft_email.body = body if body else draft_email.body or ""
    draft_email.update_recipient_lists(
        to_add=to_add,
        to_remove=to_remove,
        cc_add=cc_add,
        cc_remove=cc_remove,
        bcc_add=bcc_add,
        bcc_remove=bcc_remove,
    )
    updated_draft_email = await client.me.messages.by_message_id(message_id).patch(
        draft_email.to_sdk()
    )

    return Message.from_sdk(updated_draft_email).to_dict()  # type: ignore[arg-type]



import asyncio
import os
from dataclasses import dataclass
from typing import Any, cast

import httpx

from arcade_salesforce.constants import MAX_CONCURRENT_REQUESTS, SALESFORCE_API_VERSION
from arcade_salesforce.enums import SalesforceObject
from arcade_salesforce.exceptions import (
    BadRequestError,
    ResourceNotFoundError,
    SalesforceToolExecutionError,
)
from arcade_salesforce.utils import (
    build_soql_query,
    clean_contact_data,
    clean_lead_data,
    clean_note_data,
    clean_object_data,
    clean_opportunity_data,
    clean_task_data,
    expand_associations,
    get_ids_referenced,
    get_object_type,
    remove_none_values,
)


@dataclass
class SalesforceClient:
    auth_token: str
    org_subdomain: str | None = None
    api_version: str = SALESFORCE_API_VERSION
    max_concurrent_requests: int = MAX_CONCURRENT_REQUESTS

    # Internal state properties
    _state_object_fields: dict[SalesforceObject, list[str]] | None = None
    _state_is_person_account_enabled: bool | None = None
    _semaphore: asyncio.Semaphore | None = None

    def __post_init__(self) -> None:
        if self.org_subdomain is None:
            self.org_subdomain = os.getenv("SALESFORCE_ORG_SUBDOMAIN")
        if self.org_subdomain is None:
            raise ValueError(
                "Either `org_subdomain` argument or `SALESFORCE_ORG_SUBDOMAIN` env var must be set"
            )

        if self._state_object_fields is None:
            self._state_object_fields = {}

        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self.max_concurrent_requests)

    @property
    def _base_url(self) -> str:
        return f"https://{self.org_subdomain}.my.salesforce.com/services/data/{self.api_version}"

    @property
    def object_fields(self) -> dict[SalesforceObject, list[str]]:
        return cast(dict, self._state_object_fields)

    def _endpoint_url(self, endpoint: str) -> str:
        return f"{self._base_url}/{endpoint.lstrip('/')}"

    def _build_headers(self, headers: dict | None = None) -> dict:
        headers = headers or {}
        headers["Authorization"] = f"Bearer {self.auth_token}"
        return headers

    def _raise_salesforce_error(self, response: httpx.Response) -> None:
        """Raise a ToolExecutionError based on the Salesforce API response status code."""
        errors = [error.get("message") for error in response.json()]
        if response.status_code == 404:
            raise ResourceNotFoundError(errors)
        elif response.status_code == 400:
            raise BadRequestError(errors)
        raise SalesforceToolExecutionError(errors)

    async def get(
        self,
        endpoint: str,
        params: dict | None = None,
        headers: dict | None = None,
    ) -> dict:
        """Make a GET request to the Salesforce API.

        Args:
            endpoint: The Salesforce API endpoint to call.
            params: The query parameters to include in the request.
            headers: The headers to include in the request.

        Returns:
            The JSON-loaded representation of the response from the Salesforce API.
        """
        kwargs: dict[str, Any] = {
            "url": self._endpoint_url(endpoint),
            "headers": self._build_headers(headers),
        }
        if params:
            kwargs["params"] = params

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.get(**kwargs)

        if response.status_code >= 300:
            self._raise_salesforce_error(response)

        return cast(dict, response.json())

    async def post(
        self,
        endpoint: str,
        data: dict | None = None,
        json_data: dict | None = None,
        headers: dict | None = None,
    ) -> dict:
        """Make a POST request to the Salesforce API.

        Args:
            endpoint: The Salesforce API endpoint to call.
            data: The data for the request. (provide data or json_data, not both)
            json_data: The JSON data for the request. (provide data or json_data, not both)
            headers: The headers to include in the request.

        Returns:
            The JSON-loaded representation of the response from the Salesforce API.
        """
        kwargs: dict[str, Any] = {
            "url": self._endpoint_url(endpoint),
            "headers": self._build_headers(headers),
        }
        if data:
            kwargs["data"] = data
        if json_data:
            kwargs["json"] = json_data

        async with self._semaphore, httpx.AsyncClient() as client:  # type: ignore[union-attr]
            response = await client.post(**kwargs)

        if response.status_code >= 300:
            self._raise_salesforce_error(response)

        return cast(dict, response.json())

    async def get_object_fields(self, object_type: SalesforceObject) -> list[str]:
        """Get the fields available for a Salesforce object.

        Args:
            object_type: The Salesforce object to get the fields for.

        Returns:
            The list of fields available for a Salesforce object.
        """
        if object_type not in self.object_fields:
            response = await self._describe_object(object_type)
            self.object_fields[object_type] = [field["name"] for field in response["fields"]]

        return self.object_fields[object_type]

    async def _describe_object(self, object_type: SalesforceObject) -> dict:
        return await self.get(f"sobjects/{object_type.value}/describe/")

    async def _get_related_objects(
        self,
        child_object_type: SalesforceObject,
        parent_object_type: SalesforceObject,
        parent_object_id: str,
        limit: int | None = 10,
    ) -> list[dict]:
        """Get the objects that are associated with another Salesforce object.

        Args:
            child_object_type: The type of child object to retrieve.
            parent_object_type: The type of parent object.
            parent_object_id: The ID of the parent object.
            limit: The maximum number of related objects to retrieve.

        Returns:
            The list of related objects.
        """
        try:
            response = await self.get(
                f"sobjects/{parent_object_type.value}/{parent_object_id}/{child_object_type.plural.lower()}",
                params={"limit": limit},
            )
            return cast(list, response["records"])
        except ResourceNotFoundError:
            return []

    async def get_object_by_id(self, object_id: str) -> dict | None:
        """Get a Salesforce object by its ID.

        Args:
            object_id: The ID of the Salesforce object to retrieve.

        Returns:
            The Salesforce object.
        """
        try:
            response = await self.get(f"sobjects/{object_id}")
            return clean_object_data(response)
        except ResourceNotFoundError:
            if "User" not in object_id:
                return await self.get_object_by_id(f"User/{object_id}")
            return None

    async def get_account(self, account_id: str) -> dict[str, Any] | None:
        """Get an account by its ID.

        Args:
            account_id: The ID of the account to retrieve.

        Returns:
            The account.
        """
        try:
            return cast(dict, await self.get(f"sobjects/Account/{account_id}"))
        except ResourceNotFoundError:
            return None

    async def get_account_contacts(self, account_id: str, limit: int | None = 10) -> list[dict]:
        """Get the contacts associated with an account.

        Args:
            account_id: The ID of the account to retrieve the contacts for.
            limit: The maximum number of contacts to retrieve.

        Returns:
            The list of contacts with cleaned and standardized dictionaries.
        """
        contacts = await self._get_related_objects(
            SalesforceObject.CONTACT, SalesforceObject.ACCOUNT, account_id, limit
        )

        return [
            clean_contact_data(contact)
            for contact in await asyncio.gather(*[
                self.enrich_contact_with_notes(contact, limit) for contact in contacts
            ])
        ]

    async def enrich_contact_with_notes(self, contact: dict, limit: int | None = 10) -> dict:
        """Get the notes associated with a contact and add to the contact dictionary.

        Args:
            contact: The contact to retrieve the notes for.
            limit: The maximum number of notes to retrieve.

        Returns:
            The contact with the notes added to the dictionary in the key `Notes`.
        """
        notes = await self.get_notes(contact["Id"], limit)
        if notes:
            contact["Notes"] = notes
        return contact

    async def get_account_leads(self, account_id: str, limit: int | None = 10) -> list[dict]:
        """Get the leads associated with an account.

        Args:
            account_id: The ID of the account to retrieve the leads for.
            limit: The maximum number of leads to retrieve.

        Returns:
            The list of leads with cleaned and standardized dictionaries.
        """
        leads = await self._get_related_objects(
            SalesforceObject.LEAD, SalesforceObject.ACCOUNT, account_id, limit
        )
        return [clean_lead_data(lead) for lead in leads]

    async def get_notes(self, parent_id: str, limit: int | None = 10) -> list[dict]:
        """Get the notes associated with a Salesforce object.

        Args:
            parent_id: The ID of the Salesforce object to retrieve the notes for.
            limit: The maximum number of notes to retrieve.

        Returns:
            The list of notes with cleaned and standardized dictionaries.
        """
        query = build_soql_query(
            "SELECT Id, Title, Body, OwnerId, CreatedById, CreatedDate "
            "FROM Note "
            "WHERE ParentId = '{parent_id}' "
            "LIMIT {limit}",
            parent_id=parent_id,
            limit=limit,
        )
        response = await self.get("query", params={"q": query})
        notes = response["records"]
        return [clean_note_data(note) for note in notes]

    # TODO: Add support for retrieving Currency, when enabled in the org account.
    # If not enabled and we try to retrieve it, we get a 400 error.
    # More inf: https://developer.salesforce.com/docs/atlas.en-us.254.0.object_reference.meta/object_reference/sforce_api_objects_opportunity.htm#i1455437
    async def get_account_opportunities(
        self,
        account_id: str,
        limit: int | None = 10,
    ) -> list[dict]:
        """Get the opportunities associated with an account.

        Args:
            account_id: The ID of the account to retrieve the opportunities for.
            limit: The maximum number of opportunities to retrieve.

        Returns:
            The list of opportunities with cleaned and standardized dictionaries.
        """
        query = build_soql_query(
            "SELECT Id, Name, Type, StageName, OwnerId, CreatedById, LastModifiedById, "
            "Description, Amount, Probability, ExpectedRevenue, CloseDate, ContactId "
            "FROM Opportunity "
            "WHERE AccountId = '{account_id}' "
            "LIMIT {limit}",
            account_id=account_id,
            limit=limit,
        )
        response = await self.get("query", params={"q": query})
        opportunities = response["records"]
        return [clean_opportunity_data(opportunity) for opportunity in opportunities]

    async def get_account_tasks(
        self,
        account_id: str,
        limit: int | None = 10,
    ) -> list[dict]:
        """Get the tasks associated with an account.

        Args:
            account_id: The ID of the account to retrieve the tasks for.
            limit: The maximum number of tasks to retrieve.

        Returns:
            The list of tasks with cleaned and standardized dictionaries.
        """
        tasks = await self._get_related_objects(
            SalesforceObject.TASK, SalesforceObject.ACCOUNT, account_id, limit
        )
        return [clean_task_data(task) for task in tasks]

    async def enrich_account(
        self,
        account_id: str | None = None,
        account_data: dict[str, Any] | None = None,
        limit_per_association: int | None = 10,
    ) -> dict:
        """Enrich account dictionary with contacts, leads, opportunities, and tasks.

        Provide `account_id` or `account_data`, but not both.

        Args:
            account_id: The ID of the account to retrieve the associations for.
            account_data: The account data to enrich.
            limit_per_association: The maximum number of associations to retrieve.

        Returns:
            The account with the associations added to the dictionary.
        """
        if (account_id and account_data) or (not account_id and not account_data):
            raise ValueError("Must provide exactly one of `account_id` or `account_data`")

        if account_data is None:
            account_data = await self.get_account(cast(str, account_id))

            if not account_data:
                raise ResourceNotFoundError([f"Account not found with ID: {account_id}"])

        if not account_id:
            account_id = cast(str, account_data["Id"])

        associations = await asyncio.gather(
            self.get_account_contacts(account_id, limit=limit_per_association),
            self.get_account_leads(account_id, limit=limit_per_association),
            self.get_account_opportunities(account_id, limit=limit_per_association),
            self.get_account_tasks(account_id, limit=limit_per_association),
        )

        for association in associations:
            for item in association:
                try:
                    obj_type = SalesforceObject(get_object_type(item)).plural
                except ValueError:
                    obj_type = get_object_type(item) + "s"

                if obj_type not in account_data:
                    account_data[obj_type] = []
                account_data[obj_type].append(item)

        return await self.expand_account_associations(account_data)

    async def expand_account_associations(self, account: dict) -> dict:
        """Expand with metadata about objects referenced by ID in an account dictionary.

        This method will, for example, expand an `OwnerId` or `ContactId` referenced in an account
        dictionary with the owner or contact name, for example.

        Args:
            account: The account dictionary to expand.

        Returns:
            The account with the associations expanded.
        """
        objects_by_id = {
            obj["Id"]: obj
            for obj_type in SalesforceObject
            for obj in account.get(obj_type.plural, [])
        }
        objects_by_id[account["Id"]] = account

        referenced_ids = get_ids_referenced(
            account,
            *[account.get(obj_type.plural, []) for obj_type in SalesforceObject],
        )

        missing_referenced_ids = [ref for ref in referenced_ids if ref not in objects_by_id]

        if missing_referenced_ids:
            missing_objects = await asyncio.gather(*[
                self.get_object_by_id(missing_id) for missing_id in missing_referenced_ids
            ])
            objects_by_id.update({obj["Id"]: obj for obj in missing_objects if obj is not None})

        account = expand_associations(account, objects_by_id)

        for object_type in SalesforceObject:
            if object_type.plural not in account:
                continue

            expanded_items = []

            for item in account[object_type.plural]:
                if "AccountId" in item:
                    del item["AccountId"]

                expanded_items.append(expand_associations(item, objects_by_id))

            if object_type == SalesforceObject.CONTACT:
                for contact in expanded_items:
                    if "Notes" in contact:
                        contact["Notes"] = [
                            expand_associations(note, objects_by_id) for note in contact["Notes"]
                        ]

            account[object_type.plural] = expanded_items

        return account

    async def create_contact(
        self,
        account_id: str,
        last_name: str,
        first_name: str | None = None,
        email: str | None = None,
        phone: str | None = None,
        mobile_phone: str | None = None,
        title: str | None = None,
        department: str | None = None,
        description: str | None = None,
    ) -> dict:
        """Create a contact in Salesforce.

        Args:
            account_id: The ID of the account to associate the contact with.
            last_name: The last name of the contact.
            first_name: The first name of the contact.
            email: The email of the contact.
            phone: The phone number of the contact.
            mobile_phone: The mobile phone number of the contact.
            title: The title of the contact.
            department: The department of the contact.
            description: The description of the contact.

        Returns:
            The created contact.
        """
        data = {
            "AccountId": account_id,
            "FirstName": first_name,
            "LastName": last_name,
            "Email": email,
            "Phone": phone,
            "MobilePhone": mobile_phone,
            "Title": title,
            "Department": department,
            "Description": description,
        }

        return await self.post(
            f"sobjects/{SalesforceObject.CONTACT.value}",
            json_data=remove_none_values(data),
        )



from arcade_evals import (
    BinaryCritic,
    EvalRubric,
    EvalSuite,
    ExpectedToolCall,
    tool_eval,
)
from arcade_tdk import ToolCatalog

import arcade_confluence
from arcade_confluence.tools import (
    get_space,
    get_space_hierarchy,
    list_spaces,
)
from evals.conversations import list_spaces_1

# Evaluation rubric
rubric = EvalRubric(
    fail_threshold=0.85,
    warn_threshold=0.95,
)


catalog = ToolCatalog()
catalog.add_module(arcade_confluence)


@tool_eval()
def confluence_get_space_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence get_space tool evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get two spaces - one by ID, and one by key",
        user_message="Get spaces 3498573 and 'Poetry'",
        expected_tool_calls=[
            ExpectedToolCall(func=get_space, args={"space_identifier": 3498573}),
            ExpectedToolCall(func=get_space, args={"space_identifier": "Poetry"}),
        ],
        rubric=rubric,
        critics=[BinaryCritic(critic_field="space_identifier", weight=1)],
    )

    return suite


@tool_eval()
def confluence_list_spaces_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence list_spaces tool evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="List the next space using a pagination token",
        user_message="get the next one",
        expected_tool_calls=[
            ExpectedToolCall(
                func=list_spaces,
                args={
                    "limit": 1,
                    "pagination_token": "eyJpZCI6MjMyNjc5NCwic3BhY2VTb3J0T3JkZXIiOnsiZmllbGQiOiJOQU1FIiwiZGlyZWN0aW9uIjoiQVNDRU5ESU5HIn0sInNwYWNlU29ydE9yZGVyVmFsdWUiOiJlcmljYXJjYWRlIn0=",  # noqa: E501
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="limit", weight=0.3),
            BinaryCritic(critic_field="pagination_token", weight=0.7),
        ],
        additional_messages=list_spaces_1,
    )

    return suite


@tool_eval()
def confluence_get_space_hierarchy_eval_suite() -> EvalSuite:
    suite = EvalSuite(
        name="Confluence get_space_hierarchy tool evaluation",
        system_message="You are an AI assistant with access to Confluence tools.",
        catalog=catalog,
        rubric=rubric,
    )

    suite.add_case(
        name="Get the hierarchy of a space",
        user_message=(
            "What is the best file location within my 'Poetry' space to create a new page "
            "named 'Rough Draft - Poem for King Henry VIII'?"
        ),
        expected_tool_calls=[
            ExpectedToolCall(
                func=get_space_hierarchy,
                args={
                    "space_identifier": "Poetry",
                },
            ),
        ],
        rubric=rubric,
        critics=[
            BinaryCritic(critic_field="space_identifier", weight=1),
        ],
    )

    return suite
